c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine celij(j1,j2,kind_p1,ini_kind_p2,lx2,lz2,id_call)
c
      include 'common.2D'  

      if(j2.lt.0)then
        print*
        print*,'ERROR in celij.f'
        print*,'j2.lt.0'
        print*,'j1 ',j1
        print*,'j2 ',j2
        print*,'id_call ',id_call
        print*,'rank_id ',rank_id
        stop
      endif
      
      do kind_p2=ini_kind_p2,2
        if(nc(j2,kind_p2).ne.0) then

        do ii=1,nc(j1,kind_p1)
          i = ibox(j1,kind_p1,ii)
         
          rhou_i = Volrhou(i)/pVol(i)
          rhow_i = Volrhow(i)/pVol(i)
          
          
          do jj=1,nc(j2,kind_p2)
           j = ibox(j2,kind_p2,jj)
            
           drx = xp(i) - xp(j)
           drz = zp(i) - zp(j)
            

           call periodicityCorrection(i,j,drx,drz,lx2,lz2,id_call)


           rr2 = drx*drx + drz*drz

           if(rr2.lt.rr2_min)then
             if(i_particleType(i).lt.4.and.i_particleType(j).lt.4)then
               goto 641
             end if
              print*
              print*,'ERROR In celij.f: rr2.lt.rr2_min'
              print*,'itime ',itime
              print*,'rank_id ',rank_id
              print*,'j1, j2 ',j1, j2
              print*,'kind_p1, kind_p2 ',kind_p1, kind_p2
              print*,'I, j ',I, j
              print*,'xp(I), zp(I) ',xp(i), zp(i)
              print*,'xp(j), zp(j) ',xp(j), zp(j)
              print*,'drx, drz ',drx, drz
              print*,'rr2     ',rr2
              print*,'rr2_min ',rr2_min
              print*,'fourh2 ',fourh2
              print*,'q=sqrt(rr2)/h =',sqrt(rr2)/h
              print*,'i_particleType(i) ',i_particleType(i)
              print*,'i_particleType(j) ',i_particleType(j)
              print*,'pm(I) ',pm(I)
              print*,'pm(j) ',pm(j)
              print*,'up(I), wp(I) ',up(i), wp(i)
              print*,'up(j), wp(j) ',up(j), wp(j)
              print*,'rhop(I), rdot(I) ',rhop(i), rdot(i)
              print*,'rhop(j), rdot(j) ',rhop(j), rdot(j)
              call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
              stop
           end if
           
           if(rr2.lt.fourh2.and.rr2.gt.1.e-18) then
           
             dux = up(i) - up(j)
             duz = wp(i) - wp(j)


c            Calculating kernel & Normalized Kernel Gradient
             call kernel(drx,drz,i,j,j1,j2,rr2) 
             call kernel_correction(i,j)
             
c
c            ...  average density
c

             robar  = 0.5*( rhop(i) + rhop(j) )
             one_over_rhobar = 2.0/(rhop(i) + rhop(j))
c
             cbar = 0.5*( cs(i)   + cs(j)   ) 
c
c            ...  inner product rv
c
             dot = drx*dux + drz*duz

c	      Used to calculate the time step due to viscosity

             visc_dt=max(dot/(rr2 + eta2),visc_dt) 
c
c            ...  pressure and viscous force (1992; 3.3)
c
c            pm(j) is mass of particle j
c
c             p_v = pr(i) + pr(j)  !+  pi_visc



c           for tensile correction (Monaghan , JCP.  159 (2000) 290- 311)
c           Only to be activated with cubic spline kernel 
c
            if(index_tensile.eq.1) then  

c   _____ Tensile correction

              fab=Wab*od_Wdeltap    !NOTE: We'll use a non-normalized
              fab=fab*fab           !kernel to calculate tensile correction
              fab=fab*fab           !It's the (Wab/Wdeltap)**4  of Monaghan's paper

              if (p(i).gt.0) then
                  Ra= 0.01 *pr(i)
              else
                  Ra= 0.2 *abs(pr(i))
              endif

              if (p(j).gt.0) then
                  Rb= 0.01 *pr(j)
              else
                  Rb= 0.2 *abs(pr(j))
              endif

c              p_v = p_v+ (Ra+Rb)*fab

            endif

            if(i_particleType(i).eq.4.and.i_particleType(j).eq.4)then !FP-FP, both fluid particles

              drho    = rhop(i) - rhop(j)
              drhou   = rhou_i - (Volrhou(j)/pVol(j))   != (Volrhou(i)/pVol(i)) - (Volrhou(j)/pVol(j))
              drhow   = rhow_i - (Volrhow(j)/pVol(j))   != (Volrhow(i)/pVol(i)) - (Volrhow(j)/pVol(j))
               
              dTotalE = TotalE(i) - TotalE(j)
              
              call approx_RiemannSolver(i,j,drx,drz,rr2,
     &                           drho,drhou,drhow,dTotalE,
     &                    u_a,w_a,rho_a,p_a,TotalE_a,u_0,w_0)
              p_v_x1 = rho_a*u_a*(u_a - u_0) + p_a
c             p_v_x2
              p_v_x3 = rho_a*w_a*(u_a - u_0)
c             p_v_y1
c             p_v_y2
c             p_v_y3
              p_v_z1 = rho_a*u_a*(w_a - w_0)
c             p_v_z2
              p_v_z3 = rho_a*w_a*(w_a - w_0) + p_a
              frx_tmpj = pVol(j)*2.0*frxi
              frz_tmpj = pVol(j)*2.0*frzi
              
              frx_tmpi = pVol(i)*2.0*frxj
              frz_tmpi = pVol(i)*2.0*frzj
              
              ax(i) = ax(i)
     &            -(p_v_x1*frx_tmpj + p_v_x3*frz_tmpj)
c             ay(i) = ay(i)
c    &            -(                                 )
              az(i) = az(i)
     &            -(p_v_z1*frx_tmpj + p_v_z3*frz_tmpj)
              ax(j) = ax(j)
     &            +(p_v_x1*frx_tmpi + p_v_x3*frz_tmpi)
c             ay(j) = ay(j)
c    &            -(                                 )
              az(j) = az(j)
     &            +(p_v_z1*frx_tmpi + p_v_z3*frz_tmpi)

c              i_particleCheck = -265431
c              if(i.eq.i_particleCheck.and.itime.eq.00)then
c                d_az = -(p_v_z1*frx_tmpj + p_v_z3*frz_tmpj)
c              end if
c              if(j.eq.i_particleCheck.and.itime.eq.00)then   !.and.itime.eq.00
c                d_az =  (p_v_z1*frx_tmpi + p_v_z3*frz_tmpi)
c              end if
c              i_particleCheck = 537  !-265431
c              j_particleCheck = 675  !265454
c              if(i.eq.i_particleCheck.and.itime.eq.00)then
c               print*
c                print*,'In celij_BC_Monaghan_2D.f'
c                print*,'i ',i
c                print*,'xp, zp  ',xp(i),zp(i)
c                print*,'rhop, p ',rhop(i),p(i)
c                print*,'up, TotalE ',up(i),TotalE(i)
c                print*,'j ',j
c                print*,'xp, zp ',xp(j),zp(j)
c                print*,'rhop, p    ',rhop(j),p(j)
c                print*,'up, TotalE ',up(j),TotalE(j)
c                print*,'u_a, w_a ',u_a, w_a
c                print*,'u_0, w_0 ',u_0, w_0
c                print*,'rho_a ',rho_a
c                print*,'p_a   ',p_a
c                print*,'TotalE_a   ',TotalE_a
c                print*,'ax(i) ',ax(i)
c                print*,'az(i) ',az(i)
c                print*,'-(p_v_x1*frx_tmpj + p_v_x3*frz_tmpj)',
c     &                  -(p_v_x1*frx_tmpj + p_v_x3*frz_tmpj)           
c                print*,'-(p_v_z1*frx_tmpj + p_v_z3*frz_tmpj)',
c     &                  -(p_v_z1*frx_tmpj + p_v_z3*frz_tmpj)           
c                print*,'p_v_x1   ',p_v_x1
c                print*,'p_v_x3   ',p_v_x3
c                print*,'p_v_z1   ',p_v_z1
c                print*,'p_v_z3   ',p_v_z3
c                print*,'frx_tmpj ',frx_tmpj           
c                print*,'frz_tmpj ',frz_tmpj           
c                print*,'frx     ',frx           
c                print*,'drx , drz ',drx,drz
c                print*,'pVol(j) ',pVol(j)
c                print*,'dot2aj = (u_a - u_0)*frxz_tmpj ',
c     &                   (u_a - u_0)*frx_tmpj + (w_a - w_0)*frz_tmpj
c                print*,'- rho_a*dot2aj   ',
c     &          - rho_a*((u_a - u_0)*frx_tmpj + (w_a - w_0)*frz_tmpj)
c                print*,'ar(i)         ',
c     &                  ar(i) - rho_a*(u_a - u_0)*frx_tmpj
c                !stop
c              end if
c              if(j.eq.i_particleCheck.and.itime.eq.00)then   !.and.itime.eq.00
c                print*
c                print*,'In celij_BC_Monaghan_2D.f'
c                print*,'j ',j
c                print*,'xp, zp ',xp(j),zp(j)
c                print*,'rhop, p    ',rhop(j),p(j)
c                print*,'up, TotalE ',up(j),TotalE(j)
c                print*,'i ',i
c                print*,'xp, zp ',xp(i),zp(i)
c                print*,'rhop, p    ',rhop(i),p(i)
c                print*,'up, TotalE ',up(i),TotalE(i)
c                print*,'u_a, w_a ',u_a, w_a
c                print*,'u_0, w_0 ',u_0, w_0
c                print*,'rho_a ',rho_a
c                print*,'p_a   ',p_a
c                print*,'TotalE_a   ',TotalE_a
c                print*,'ax(j) ',ax(j)
c                print*,'az(j) ',az(j)
c                print*,'+(p_v_x1*frx_tmpi + p_v_x3*frz_tmpi)',
c     &                   (p_v_x1*frx_tmpi + p_v_x3*frz_tmpi)           
c                print*,'+(p_v_z1*frx_tmpi + p_v_z3*frz_tmpi)',
c     &                   (p_v_z1*frx_tmpi + p_v_z3*frz_tmpi)           
c                print*,'p_v_x1   ',p_v_x1
c                print*,'p_v_x3   ',p_v_x3
c                print*,'p_v_z1   ',p_v_z1
c                print*,'p_v_z3   ',p_v_z3
c                print*,'frx_tmpi ',frx_tmpi           
c                print*,'frz_tmpi ',frz_tmpi           
c                print*,'frx     ',frx           
c                print*,'-drx,-drz ',-drx,-drz
c                print*,'pVol(i) ',pVol(i)
c                print*,'dot2ai = (u_a - u_0)*frxz_tmpi',
c     &                 (u_a - u_0)*frx_tmpi + (w_a - w_0)*frz_tmpi
c                print*,'+ rho_a*dot2ai   ',
c     &          rho_a*((u_a - u_0)*frx_tmpi + (w_a - w_0)*frz_tmpi)
c                print*,'ar(j)         ',
c     &                  ar(j) + rho_a*(u_a - u_0)*frx_tmpi
c                !stop
c              end if

              call gradients_calc(i,j,drho,drhou,drhow,dTotalE)                                             
               
              call viscosity(dot,drx,drz,dux,duz,rr2,
     +               cbar,robar,one_over_rhobar,i,j,j1,j2,term2i,term2j)

c  ...  density acceleration
c 
              dot2ai = (u_a - u_0)*frx_tmpi
     &               + (w_a - w_0)*frz_tmpi
              dot2aj = (u_a - u_0)*frx_tmpj
     &               + (w_a - w_0)*frz_tmpj
              ar(i) = ar(i) - rho_a*dot2aj
              ar(j) = ar(j) + rho_a*dot2ai
              dot2i = dux*frxi + duz*frzi
              dot2j = dux*frxj + duz*frzj
              aVol(i) = aVol(i) - dot2i*pVol(j)
              aVol(j) = aVol(j) - dot2j*pVol(i)

c  ...  Thermal Energy
c
              aTE(i) = aTE(i) 
     &           - (TotalE_a*dot2aj
     &              + p_a*(u_a*frx_tmpj + w_a*frz_tmpj))
              aTE(j) = aTE(j) 
     &           + (TotalE_a*dot2ai
     &              + p_a*(u_a*frx_tmpi + w_a*frz_tmpi))
c
c  ...  XSPH correction (Monaghan 1994;  J. Comp. Phys. 110. Formula 2.6)
c
              pmj_Wab_over_rhobar = pm(j)*Wab*one_over_rhobar
              ux(i) = ux(i) - dux*pmj_Wab_over_rhobar  !pm(j) * dux * Wab / robar
              wx(i) = wx(i) - duz*pmj_Wab_over_rhobar  !pm(j) * duz * Wab / robar


              pmi_Wab_over_rhobar = pm(i)*Wab*one_over_rhobar
              ux(j) = ux(j) + dux*pmi_Wab_over_rhobar   !pm(i) * dux * Wab / robar
              wx(j) = wx(j) + duz*pmi_Wab_over_rhobar   !pm(i) * duz * Wab / robar


c  ...  Vorticity calculation

              if(ipoute.eq.1.and.i_vort.eq.1.and.
     &          i_particleType(i).gt.3.and.i_particleType(j).gt.3)then  !FP-FP
                  call vorticity_calc(i,j,dux,duz)
              endif  
                                   
            else if(i_particleType(i).eq.4.and.
     &              i_particleType(j).lt.4)then  !FP-BP,   j is boundary particle
              call MonaghanBC(i,j,drx,drz,dot,dux,duz,
     &                        fxbMON,fzbMON)   
              ax(i) = ax(i) + fxbMON*rhop(i)
              az(i) = az(i) + fzbMON*rhop(i)
              
c             temp = pVol(i)*visc_wall*
c     &          ((drx*frxi+drz*frzi)/(rr2 + eta2))
c             ax(i) = ax(i) + temp*dux
c             az(i) = az(i) + temp*duz

            
            else if(i_particleType(j).eq.4.and.
     &              i_particleType(i).lt.4)then  !FP-BP,   i is boundary particle
              call MonaghanBC(j,i,-drx,-drz,dot,-dux,-duz,
     &                        fxbMON,fzbMON)   
              ax(j) = ax(j) + fxbMON*rhop(j)
              az(j) = az(j) + fzbMON*rhop(j)
               
c             temp = pVol(j)*visc_wall*
c     &          ((drx*frxj+drz*frzj)/(rr2 + eta2))
c             ax(j) = ax(j) - temp*dux
c             az(j) = az(j) - temp*duz


             endif  ! Interaction fluid-fluid or fluid-boundary

           endif ! if q<2
          enddo ! Box jj
         enddo ! Box ii
641     endif  ! Box jj is not empty
      enddo   ! Kind of particle

      end

		  
c           !-- Debug Code------
c           if(rank_id.eq.0)then
c             i_particleCheck = -265431
c             if(i.eq.i_particleCheck)then   !if(i.eq.5661)then
c                i_drx = nint(drx/(vnorm_mass*dpx))
c                i_drz = nint(drz/(vnorm_mass*dpz))
c                write(*,857)'I,j, j1, j2, drx, drz, az  ',
c     &                  i,j, j1, j2, i_drx, i_drz, az(i), d_az 
c             elseif(j.eq.i_particleCheck)then   !elseif(j.eq.5661)then
c                i_drx = nint(drx/(vnorm_mass*dpx))
c                i_drz = nint(drz/(vnorm_mass*dpz))
c                write(*,857)'i,J, j1, j2,-drx,-drz, az  ',
c     &                  i,J, j1, j2,-i_drx,-i_drz, az(j), d_az 
c             endif           
c           endif
c857        format(a23,2i8,2i6,2i3,2(' ',e24.16))
c           !-------------------
