c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine divide(n_start,n_end,kind_p,i_resident)

      include 'common.2D'


c     i_resident = 1   =>  Resident particles
c     i_resident = 2   =>  Guest particles 

c     icell = 1 & icell = ncx_local : reserved for Guest particles only
c
c     ist and ien are the starting and ending values of the positions in the linear
c     array ip (corresponding to a list of particle numbers) for the particles 
c     in cell ii
c

      itime_check = 0   !10e8   !0   !10e8   !0
      
      !---  CHECK THIS IS NOT NEEDED SINCE WE NOW HAVE ini_divide, keep_list, recover_list ---!!!!!!!!!
      if(i_resident.eq.1.or.i_resident.eq.0)then  !Zero nc array only if rewriting nc for resident particles
        !- NOW IN INI_DIVIDE -
        !do i=1,nct_local
        !  ist(i,kind_p) = 1
        !  ien(i,kind_p) = 0
        !  nc(i,kind_p)  = 0
        !  ip(i,kind_p)  = 0
        !enddo
        nb_temp_local        = nb_local
        nbf_temp_local       = nbf_local
        nbfm_temp_local      = nbfm_local
        nbfreem_temp_local   = nbfreem_local
      else
        nb_temp_local        = nb_image_local
        nbf_temp_local       = nbf_image_local
        nbfm_temp_local      = nbfm_image_local
        nbfreem_temp_local   = nbfreem_image_local
      end if

      
      nc_ii_max = 0
      
      do kpcount=n_start,n_end	
      
        !if(kpcount.le.nb_temp_local)then
        !  k = ip_index_BP_local(kpcount)
        !else
        !  k = ip_index_FP_local(kpcount-nb_temp_local)
        !end if
        if(kpcount.le.nbf_temp_local)then
            k = ip_index_BP_local_1(kpcount)
        else if(kpcount.le.nbfm_temp_local)then
            k = ip_index_BP_local_2(kpcount-nbf_temp_local)
        else if(kpcount.le.nbfreem_temp_local)then
            k = ip_index_BP_local_3(kpcount-nbfm_temp_local)
        else
            k = ip_index_FP_local(kpcount-nb_temp_local)
        end if
        
         !-- Error Catching --
         if(k.lt.1)then
           print*
           print*,'ERROR in divide.f'
           print*,'k.lt.1'
           print*,'kpcount, k ',kpcount, k
           print*,'itime ',itime
           print*,'kind_p ',kind_p
           print*,'rank_id, rank_idx, rank_idz ',
     &             rank_id, rank_idx, rank_idz
           print*,'n_start, n_end ',n_start,n_end
           print*,'i_resident ',i_resident
           print*,'nb_local ',nb_local
           print*,'np_local ',np_local
           print*,'nbf_image_local       = ',nbf_image_local
           print*,'nbfm_image_local = ',nbfm_image_local
           print*,'nbfreem_image_local     = ',nbfreem_image_local
           print*,'nb_image_local ',nb_image_local
           print*,'np_image_local ',np_image_local
           print*
           call MPI_ABORT(MPI_COMM_WORLD, itemp, ierr)
           stop
         end if
        
        if (iflag(k).ne.0) then	
          dx = xp(k) - xmin_local
          dz = zp(k) - zmin_local

          icell = int( dx * one_over_2h ) + 1
          kcell = int( dz * one_over_2h ) + 1

         !-- Correcting for particles being in the wrong (ghost) columns --
         if(i_resident.lt.2)then   !- Correction for resident particles being placed in ghost columns-
           !-  X-direction   -
           if(icell.gt.ncx_local-1)then
             icell = ncx_local-1  !Place back in previous column 
                                  !(avoids particle moving to next rank prematurely)
                                  !see getdata.f 
           elseif(icell.lt.2)then
             icell = 2            !Place P in next column 
                                  !(avoids particle moving to previous rank prematurely)
                                  !see getdata.f
           endif
           !-  Z-direction   -
           if(kcell.gt.ncz_local-1)then
             kcell = ncz_local-1  !Place back in previous column 
                                  !(avoids particle moving to next rank prematurely)
                                  !see getdata.f 
           elseif(kcell.lt.2)then
             kcell = 2            !Place P in next column 
                                  !(avoids particle moving to previous rank prematurely)
                                  !see getdata.f
           endif
         else    !i_resident.eq.2   !- Correction for guest particles not being ghost columns-
           if(icell.eq.ncx_local-1.or.icell.eq.ncx_local+1)then
             if(xp(k).gt.xmax_rank)then
               icell = ncx_local  !Place back in ghost column 
             endif
           elseif(icell.eq.0)then
             icell = 1          !Place P in ghost column 
           elseif(icell.eq.2)then
             if(ndim.lt.2)then
               icell = 1          !Place P in ghost column 
             elseif(zp(k).lt.zmax_rank)then
               icell = 1
             endif
           endif
           !--- MPI-PERIODICITY CORRECTION, X-Direction ---
           if(iCartTop_periodicity(1))then
             !-- Special correction n_procs_x = 1 --
             if(n_procs_x.eq.1)then
               if(icell.eq.1)then
                 icell = ncx_local
               else if(icell.eq.ncx_local)then
                 icell = 1
               end if
             else  !-- Periodic MPI overlap region for last column --
               if(rank_idx.eq.0.and.icell.gt.ncx_local)then  !This .gt. should be .ge. ????
                 icell = 1
               else if(rank_idx.eq.n_procs_x-1.and.icell.lt.1)then  !And this .le. should be .ge. ????
                 icell = ncx_local  !Puts Particle in last column for periodicity
               end if
             end if
           end if

           if(kcell.eq.ncz_local-1.or.kcell.eq.ncz_local+1)then
             if(zp(k).gt.zmax_rank)then
               kcell = ncz_local  !Place back in ghost sheet              
             endif
           elseif(kcell.eq.0)then
             kcell = 1          !Place P in ghost sheet 
           elseif(kcell.eq.2)then
             if(zp(k).lt.zmin_rank)then
               kcell = 1          !Place P in ghost sheet 
             endif
           endif
           !--- MPI-PERIODICITY CORRECTION, Z-Direction ---
           if(iCartTop_periodicity(3))then
             !-- Special correction n_procs_z = 1 --
             if(n_procs_z.eq.1)then
               if(kcell.eq.1)then
                 kcell = ncz_local
               else if(kcell.eq.ncz_local)then
                 kcell = 1
               end if
             else  !-- Periodic MPI overlap region for last sheet --
               if(rank_idz.eq.0.and.kcell.gt.ncz_local)then  !This .gt. should be .ge. ????
                 kcell = 1
               else if(rank_idz.eq.n_procs_z-1.and.kcell.lt.1)then  !And this .le. should be .ge. ????
                 kcell = ncz_local  !Puts Particle in last sheet for periodicity
               end if
             end if
           end if

         endif

         !--- Periodic OBs correction ---
         !activated when not using cartesian topology in direction 1-3
         !if(i_periodicOBs(1).eq.1.and.icell.gt.ncx_local)icell=ncx_local  !Periodic Open Boundary correction X-Direction   
         !if(i_periodicOBs(3).eq.1.and.kcell.gt.ncz_local)kcell=ncz_local  !Periodic Open Boundary correction Z-Direction   
         !if(jcell.ge.ncy_local) jcell = ncy_local

         ! ii is the linear cell position in the matrix of 2h cells
         ii = icell + (kcell - 1)*ncx_local

         !-- Error Catching --
         if(ii.gt.nct_local.or.ii.lt.1)then
           print*
           print*,'ERROR in divide.f'
           if(ii.gt.nct_local)then
             print*,'Box ii.gt.nct_local'
             print*,'Do you need to RESET zmax??'
           endif
           if(ii.lt.1        )print*,'Box ii.lt.1'
           print*,'k',k
           print*,'i_particleType(k) ',i_particleType(k)
           if(i_particleType(k).lt.4)then
             kkp_absolute = iabsolute_BP_Pointer_Index(k)
             print*,'kkp_absolute ',kkp_absolute
             if(i_particleType(k).eq.3)then
               k_num_FB = i_FB_Pointer_Info(kkp_absolute)
               print*,'k_num_FB ',k_num_FB
             endif
           endif
           print*,'xp(k)',xp(k)
           print*,'zp(k)',zp(k)
           print*,'rhop(k)',rhop(k)
           print*,'dx ',dx
           print*,'dz ',dz
           print*,'xmin_local ',xmin_local
           print*,'zmin_local ',zmin_local
           print*,'xmax_local ',xmax_local
           print*,'zmax_local ',zmax_local
           print*,'xmin_rank ',xmin_rank
           print*,'xmax_rank ',xmax_rank
           print*,'zmin_rank ',zmin_rank
           print*,'zmax_rank ',zmax_rank
           print*,'itime ',itime
           print*,'ii',ii
           print*,'icell',icell
           print*,'kcell',kcell
           print*,'kind_p ',kind_p
           print*,'rank_id, rank_idx, rank_idz ',
     &             rank_id, rank_idx, rank_idz
           print*,'n_start, n_end ',n_start,n_end
           print*,'i_resident ',i_resident
           print*,'ncx_local ',ncx_local
           print*,'ncz_local ',ncz_local
           print*,'nct_local ',nct_local
           print*,'int( dx * one_over_2h ) + 1 ',int(dx*one_over_2h)+1
           print*,'int( dz * one_over_2h ) + 1 ',int(dz*one_over_2h)+1
           print*
           stop
         end if
          
          ! nc is the number of particles in cell ii
          nc(ii,kind_p) = nc(ii,kind_p)+1	
          if(nc(ii,kind_p).gt.nc_ii_max) nc_ii_max = nc(ii,kind_p) 

!           !-- Diagnostic Information --          
!           if(rank_id.eq.0)then
!            !if((k.eq.53354.or.k.eq.66829).and.itime.eq.0)then
!            if((k.eq.23093.or.k.eq.-3496).and.
!      &         itime.eq.itime_check)then
!              print*
!              print*,'kpcount, k ',kpcount,k
!              print*,'i_particleType(k) ',i_particleType(k)
!              print*,'xp(k), zp(k)',xp(k), zp(k)
!              print*,'icell,kcell',icell,kcell
!              print*,'box ii',ii
!              print*,'nc(ii,kind_p) ',nc(ii,kind_p)
!              print*,'rank_id, rank_idx, rank_idz ',
!      &               rank_id, rank_idx, rank_idz
!              print*,'xmin_local, zmin_local ',
!      &               xmin_local, zmin_local 
!              print*,'ncx_local , ncz_local  ',
!      &               ncx_local , ncz_local 
!              print*,'kind_p ',kind_p
!              print*,'itime, i_resident ',itime,i_resident
!              print*,'n_start, n_end ',n_start,n_end
!              print*,'nbfm_temp_local ',nbfm_temp_local
!              print*,'np_image_local       ',np_image_local
!              print*,'one_over_2h ',one_over_2h
!              print*,'dx, dz',dx, dz
!              print*,'dx *one_over_2h  ',dx *one_over_2h
!              print*,'int(dx * one_over_2h)  ',int(dx * one_over_2h)
!              !read(*,*)
!              !stop
!            end if
!          end if

          
          if(nc(ii,kind_p).gt.nplink_max)then
           print*
           print*,'ERROR in divide_MPI_2D.f'
           print*,'nc(ii,kind_p) >= nplink_max'
           print*,'rank_id, itime',rank_id, itime
           print*,'rank_id, rank_idx, rank_idz ',
     &             rank_id, rank_idx, rank_idz
           print*,'nc(ii,kind_p)',nc(ii,kind_p)
           print*,'kind_p       ',kind_p
           print*,'nplink_max',nplink_max
           print*,'kpcount, k ',kpcount,k
           print*,'xp(k), zp(k)',xp(k), zp(k)
           print*,'icell, kcell ',icell, kcell
           print*,'Box ii ',ii
           print*,'n_start, n_end ',n_start,n_end
           print*,'i_resident ',i_resident
           if(rank_id.eq.3.or.rank_id.eq.0)then
             do mmkind = 1,2
              print*,'mmkind ',mmkind
              do iii = 1,nc(ii,mmkind)
                i = ibox(ii,mmkind,iii)
                print*,'iii, i,xp, zp ',iii,i,xp(i),zp(i)
              end do
             end do
           endif
           stop
          end if
          
          ibox(ii,kind_p,nc(ii,kind_p))=k  !Tells us that particle with array location k (i.e. xp(k) )  
                                         !is in box ii which, so far, contains "nc(ii,mkind)" particles
          
          
c          ien(ii,kind_p) = ien(ii,kind_p)+1
c
c          do m=ien(nct_local,kind_p),ien(ii,kind_p),-1
c            if(m.gt.0)ip(m+1,kind_p) = ip(m,kind_p)
c          enddo
c   
c          do m=ii+1,nct_local
c            ist(m,kind_p)=ist(m,kind_p)+1
c            ien(m,kind_p)=ien(m,kind_p)+1
c          enddo
c		    
c          ip(ien(ii,kind_p),kind_p) = k
                 

        endif
     
      enddo 


! c      !-- Diagnostic Printout --
!       if(rank_id.eq.-10.and.itime.ge.0)then   !.and.mkind.eq.1)then
!       !if(rank_id.eq.0.and.itime.ge.itime_check)then   !.and.mkind.eq.1)then
!         print*
!         print*,'At end of Divide, kind_p = ',kind_p
!         print*,'itime ',itime
!         print*,'rank_id = ',rank_id
!         print*,'rank_id, rank_idx, rank_idz ',
!      &          rank_id, rank_idx, rank_idz
!         print*,'i_resident   ',i_resident
!         print*,'n_start, n_end ',n_start,n_end
!         print*,'np_image_local ',np_image_local
!         print*,'nb_temp_local        ', nb_temp_local
!         print*,'nbf_temp_local       ', nbf_temp_local
!         print*,'nbfm_temp_local ', nbfm_temp_local
!         print*,'nbfreem_temp_local     ', nbfreem_temp_local
!         print*,'np_local        ',np_local
!         print*,'nb_local        ', nb_local
!         print*,'nbf_local       ', nbf_local
!         print*,'nbfm_local ', nbfm_local
!         print*,'nbfreem_local     ', nbfreem_local
!         !print*,'xmin_local, ymin_local',xmin_local, ymin_local
!         print*,'nct_local ',nct_local
!         ii = 3*ncx_local
!         print*,'Box ii ',ii
!         print*,'nc(ii,mkind=1) ',nc(ii,1)
!         print*,'nc(ii,mkind=2) ',nc(ii,2)
!         print*,'nc_ii_max',nc_ii_max
!         !print*
!           print*,'For each kind'
!           do mmkind = 1,2
!             print*,'mmkind ',mmkind
!             do iii = 1,nc(ii,mmkind)
!               i = ibox(ii,mmkind,iii)
!               print*,'iii, i,xp, zp ',iii,i,xp(i),zp(i)
!             end do
!           end do
!         print*
!         if(1.eq.1)then
!         ii = 3*ncx_local-1
!         print*,'Box ii ',ii
!         print*,'nc(ii,mkind=1) ',nc(ii,1)
!         print*,'nc(ii,mkind=2) ',nc(ii,2)
!         print*,'nc_ii_max',nc_ii_max
!         !print*
!           print*,'For each kind'
!           do mmkind = 1,2
!             print*,'mmkind ',mmkind
!             do iii = 1,nc(ii,mmkind)
!               i = ibox(ii,mmkind,iii)
!               print*,'iii, i,xp, zp ',iii,i,xp(i),zp(i)
!             end do
!           end do
!         print*
!         endif   !End of:  if(1.eq.2)then
!       end if
      if(itime.eq.-1.and.kind_p.eq.2.and.i_resident.eq.2)stop
      
      do ii = 1,nct_local 
        !-- Error Catching --
         if(nc(ii,1).lt.0.or.nc(ii,2).lt.0)then
           print*
           print*,'ERROR in divide.f'
           print*,'nc(ii).lt.0'
           print*,'ii',ii
           print*,'nc(ii,1) ',nc(ii,1)
           print*,'nc(ii,2) ',nc(ii,2)
           print*,'xmin_local ',xmin_local
           print*,'zmin_local ',zmin_local
           !print*,'icell',icell
           !print*,'jcell',jcell
           !print*,'kcell',kcell
           !print*,'mkind ',mkind
           print*,'rank_id, itime ',rank_id,itime
           print*,'rank_id, rank_idx, rank_idz ',
     &             rank_id, rank_idx, rank_idz
           print*,'n_start, n_end ',n_start,n_end
           print*,'i_resident ',i_resident
           print*,'For each kind'
           do mmkind = 1,2
            print*,'mmkind ',mmkind
            do iii = 1,nc(ii,mmkind)
              i = ibox(ii,mmkind,iii)
              print*,'iii, i,xp, zp ',iii,i,xp(i),zp(i)
            end do
           end do
           print*
           stop
         end if
       end do

      end