c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

c
c*********************************************************************
c   !BDR 04/09/06
c   This subroutine performs the following operations:
c
c     (i)   Checks to see if particles are leaving the subdomain of 
c           the processor (i.e. the rank)
c     (ii)  Transfers the particles and all their properties to the
c           adjacent process
c     (iii) On each processor, the ip index system is then reorganised
c           for the beginning of the next timestep
c
c     Note, this subroutine re-organises the index pointer system but does not
c     reorganise the data array
      
      subroutine identify_Ps_for_Exporting_toWest
      
      include 'common.2D'


c
c      itime_check = 10e8   !0   !10e8

c      if(rank_id.ge.0.and.itime.ge.itime_check)then
c        print*
c        print*,'BEGINNING OF identify_Ps_for_Exporting_toWest.f'
c        print*,'itime, rank_id ',itime, rank_id
c        print*,'nb_local ',nb_local
c        print*,'np_local ',np_local
c      end if

      nbf_in_lx_2 = 0
      nbfm_in_lx_2  = 0
      nbfreem_in_lx_2  = 0
      nb_in_lx_2   = 0
      nFP_in_lx_2  = 0
      nbf_in_lx_ncx_localminus1 = 0
      nbfm_in_lx_ncx_localminus1  = 0
      nbfreem_in_lx_ncx_localminus1  = 0
      nb_in_lx_ncx_localminus1   = 0
      nFP_in_lx_ncx_localminus1  = 0

c      if(itime.ge.itime_check)then      
c        if(rank_id.eq.-10)then
c          print*
c          print*,'In identify_Ps_for_Exporting_toWest'
c          print*,'rank_id, ncx_local, ncz_local',
c     &            rank_id, ncx_local, ncz_local
c          print*,'nbf_local, nbfreem_local, np_local',
c     &            nbf_local, nbfreem_local, np_local
c          print*,'xmin_local, xmax_local ',
c     &            xmin_local, xmax_local 
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end if
c      end if
      
      do kind1=1,2
      
      if(kind1.eq.1)then
        Ntype_check = 3   !Check BPs in order of 1=Fixed BPs; 2=Forced Moving BPs; 3=Free Moving BPs
      else
        Ntype_check = 1
      endif
      
      do i_typeCheck = 1,Ntype_check
      
      do llx = 1,1   !do llx = 1,2
        if(llx.eq.1)then
          lx = 2          !Column to be exported Westwards
        else
          lx = ncx_local-1  !Column to be exported Eastwards
        end if
      
        do lz = 1,ncz_local
      
      
          j1 = lx + (lz-1)*ncx_local
      
          if(nc(j1,kind1).gt.0)then
            
            do ii=1,nc(j1,kind1)
              i = ibox(j1,kind1,ii)  !i here is the absolute position in the array, e.g. xp(i)
                                     !i IS NOT the sequential index which points to position i
            
              !if(rank_id.eq.1)then
              !  print*
              !  print*,'rank_id, itime ',rank_id, itime
              !  print*,'kind1 ',kind1
              !  print*,'llx, lx, lz', llx, lx, lz
              !  print*,'Box j1 ',j1
              !  print*,'nc(j1,kind1) ', nc(j1,kind1)
              !  print*,'is1, ie1', is1,ie1
              !  print*,'ii, i = ip(ii,kind1)',ii,i
              !  print*
              !end if
              
              if(kind1.eq.2.or.i_particleType(i).eq.i_typeCheck)then

                if(llx.eq.1)then
                  if(i_particleType(i).lt.4)then          !Boundary Particle
                    if(i_particleType(i).eq.1)then        !Fixed BP 
                      nbf_in_lx_2 = nbf_in_lx_2 + 1
                      nb_in_lx_2 = nb_in_lx_2 + 1
                      iBP_going_West_1(nbf_in_lx_2) = i
                    else if(i_particleType(i).eq.2)then   !Forced Moving BP
                      nbfm_in_lx_2  = nbfm_in_lx_2  + 1
                      nb_in_lx_2 = nb_in_lx_2 + 1
                      iBP_going_West_2(nbfm_in_lx_2) = i
                    else if(i_particleType(i).eq.3)then    !Free moving BP
                      nbfreem_in_lx_2  = nbfreem_in_lx_2  + 1
                      nb_in_lx_2 = nb_in_lx_2 + 1
                      iBP_going_West_3(nbfreem_in_lx_2) = i
                    end if                
                    iBP_going_West(nb_in_lx_2) = i
                  else if(i_particleType(i).eq.4)then      !Fluid Particle
                    nFP_in_lx_2  = nFP_in_lx_2   + 1
                    iFP_going_West(nFP_in_lx_2) = i
                  else
                    print*
                    print*,'problem: identify_Ps_for_Exporting_toWest!'
                    print*,'i.NOT.le.nb_local.OR.i.NOT.le.np_local'
                    print*,'itime ',itime
                    print*,'rank_id, ncx_local, ncz_local',
     &                      rank_id, ncx_local, ncz_local
                    print*,'nbf_local ',nbf_local
                    print*,'nbfreem_local  ', nbfreem_local
                    print*,'nb_local   ',nb_local
                    print*,'np_local   ', np_local
                    print*,'kind1 ',kind1
                    print*,'llx, lx ', llx, lx
                    print*,'Box j1 ',j1
                    print*,'nc(j1,kind1) ', nc(j1,kind1)
                    !print*,'is1, ie1', is1,ie1
                    print*,'ii, i = ip(ii,kind1)',ii,i
                    do iii=1,nc(j1,kind1)   !is1,ie1
                      print*,'iii, i, xp(i), zp(i)',
     &                        iii, i, xp(i), zp(i)
                    end do
                    call MPI_ABORT(MPI_COMM_WORLD, ierr)
                    stop
                  end if
                else if(llx.eq.2)then 
                  if(i_particleType(i).lt.4)then          !Boundary Particle
                    if(i_particleType(i).eq.1)then        !Fixed BP
                      nbf_in_lx_ncx_localminus1=
     &                nbf_in_lx_ncx_localminus1+1
                      nb_in_lx_ncx_localminus1 = 
     &                nb_in_lx_ncx_localminus1 +1
                    else if(i_particleType(i).eq.2)then   !Forced Moving BP
                      nbfm_in_lx_ncx_localminus1 = 
     &                nbfm_in_lx_ncx_localminus1+1
                      nb_in_lx_ncx_localminus1 = 
     &                nb_in_lx_ncx_localminus1 +1
                    else if(i_particleType(i).eq.3)then   !Free moving BP
                      nbfreem_in_lx_ncx_localminus1 = 
     &                nbfreem_in_lx_ncx_localminus1+1
                      nb_in_lx_ncx_localminus1 = 
     &                nb_in_lx_ncx_localminus1 +1
                    end if
                    iBP_going_East(nb_in_lx_ncx_localminus1) = i
                  else if(i_particleType(i).eq.4)then     !Fluid Particle
                    nFP_in_lx_ncx_localminus1 = 
     &              nFP_in_lx_ncx_localminus1 +1
                    iFP_going_East(nFP_in_lx_ncx_localminus1) = i
                    !if(rank_id.eq.0)then
                    !  print*
                    !  print*,'Incrementing nFP_in_lx_ncx_localminus1' 
                    !  print*,'rank_id ',rank_id
                    !  print*,'llx, lx, ly', llx, lx, ly
                    !  print*,'ii, i = ip(ii,kind1)',ii,i
                    !  print*,'xp(i), yp(i) ',xp(i), yp(i) 
                    !end if
                  else
                    print*
                    print*,'problem: identify_Ps_for_Exporting_toWest!'
                    print*,'i.NOT.le.nb_local.OR.i.NOT.le.np_local'
                    print*,'itime ',itime
                    print*,'rank_id, ncx_local, ncz_local',
     &                      rank_id, ncx_local, ncz_local
                    print*,'nbf_local ',nbf_local
                    print*,'nbfreem_local  ', nbfreem_local
                    print*,'nb_local   ',nb_local
                    print*,'np_local   ', np_local
                    print*,'kind1 ',kind1
                    print*,'llx, lx', llx, lx
                    print*,'Box j1 ',j1
                    print*,'nc(j1,kind1) ', nc(j1,kind1)
c                    print*,'is1, ie1', is1,ie1
c                    print*,'ii, i = ip(ii,kind1)',ii,i
c                    do iii=1,nc(j1,kind1)
c                      i_i = ibox(j1,kind1,iii)   !ip(ii,kind1)
c                      print*,'iii, i_i, xp(i_i), zp(i_i) ',
c     &                        iii, i_i, xp(i_i), zp(i_i)
c                    end do
c                    print*,'ii, i = ip(ii,kind1)',ii,i
                    do iii=1,nc(j1,kind1)   !is1,ie1
                      print*,'iii, i, xp(i), zp(i)',
     &                        iii, i, xp(i), zp(i)
                    end do
                    call MPI_ABORT(MPI_COMM_WORLD, ierr)
                    stop
                  end if
                end if
                
              endif    !End of:  if(kind1.eq.2.or.i_particleType(i).eq.i_typeCheck)then  
                
            end do

          end if    !End of:  if(nc(j1,kind).gt.0)then

        end do   !End of:  do lz = 1,ncz_local


      end do
      
      enddo   !End of:  do i_typeCheck = 1,3
      
      end do
      
      !end if
      !stop
      
      nbf0_in_lx_2     = nbf_in_lx_2
      nbfm0_in_lx_2    = nbfm_in_lx_2
      nbfreem0_in_lx_2 = nbfreem_in_lx_2
      nFP0_in_lx_2     = nFP_in_lx_2

      !Make numbers absolute in reference to start of local list
      !This makes the nomenclature consistent
      nbfm_in_lx_2 = nbf_in_lx_2 + nbfm_in_lx_2
      nbfreem_in_lx_2 = nbfm_in_lx_2 + nbfreem_in_lx_2
      np_in_lx_2 = nFP_in_lx_2  + nb_in_lx_2
      nbfm_in_lx_ncx_localminus1 = nbf_in_lx_ncx_localminus1 
     &                          + nbfm_in_lx_ncx_localminus1
      nbfreem_in_lx_ncx_localminus1 = nbfm_in_lx_ncx_localminus1 
     &                            + nbfreem_in_lx_ncx_localminus1
      np_in_lx_ncx_localminus1 = nFP_in_lx_ncx_localminus1
     &                         + nb_in_lx_ncx_localminus1
      
c      if(itime.ge.itime_check)then
c        if(rank_id.eq.0)then   !if(rank_id.eq.-10)then
c          print*,'Checking Particles to be exported'
c        end if
c        do i=0,n_procs-1
c          if(i.eq.rank_id)then   !if(i.eq.-rank_id-1)then   !if(i.eq.rank_id)then
c            print*
c            print*,'rank_id ', rank_id
c            print*,'nbf_in_lx_2 ',nbf_in_lx_2
c            print*,'nbfreem_in_lx_2  ',nbfreem_in_lx_2
c            print*,'nb_in_lx_2   ',nb_in_lx_2
c            print*,'nFP_in_lx_2  ',nFP_in_lx_2
c            print*,'np_in_lx_2   ',np_in_lx_2
c            print*,'nbf_in_lx_ncx_localminus1 ',
c     &              nbf_in_lx_ncx_localminus1
c            print*,'nbfreem_in_lx_ncx_localminus1  ',
c     &              nbfreem_in_lx_ncx_localminus1
c            print*,'nb_in_lx_ncx_localminus1   ',
c     &              nb_in_lx_ncx_localminus1
c            print*,'nFP_in_lx_ncx_localminus1  ',
c     &              nFP_in_lx_ncx_localminus1
c            print*,'np_in_lx_ncx_localminus1   ',
c     &              np_in_lx_ncx_localminus1
c          end if
c          call MPI_BARRIER(grid_comm, ierr)   !Synchronise all processes to this point
c        end do
c      end if
      
      !stop
      
      nb_going_East = nb_in_lx_ncx_localminus1
      np_going_East = np_in_lx_ncx_localminus1
      nb_going_West = nb_in_lx_2
      np_going_West = np_in_lx_2
      
      !-- Zero non-communicating processes --
      if(rank_idx.eq.0
     &   .and.(.NOT.iCartTop_periodicity(1)))then
        nb_going_West = 0
        np_going_West = 0
        if(n_procs_x.eq.1)then
          nb_going_East = 0
          np_going_East = 0
        end if
      else if(rank_idx.eq.n_procs_x-1
     &   .and.(.NOT.iCartTop_periodicity(1)))then
        nb_going_East = 0
        np_going_East = 0
        if(n_procs_x.eq.1)then
          nb_going_West = 0
          np_going_West = 0
        end if
      end if

c      if(itime.ge.itime_check)then
c        if(rank_id.eq.0)then
c          print*
c          print*,'At END of identify_Ps_for_Exporting_toWest'
c          print*,'Checking Particles to be exported'
c          print*,'rank_id, itime ',rank_id, itime
c          print*,'nb_going_West  ',nb_going_West
c          print*,'np_going_West  ',np_going_West
c          print*
c          print*,'BPs going West = ',nb_in_lx_2
c          do ii = 1,nb_going_West
c            i = iBP_going_West(ii)
c            print*,'ii, i, xp, zp ',ii, i, xp(i), zp(i)
c          end do
c          print*,'FPs going West = ',nFP_in_lx_2
c          do ii = nb_going_West+1,np_going_West
c            i = iFP_going_West(ii-nb_going_West)
c            print*,'ii, i, xp, zp ',ii, i, xp(i), zp(i)
c          end do
c        end if
c      end if
      
c      if(rank_id.ge.0.and.itime.ge.itime_check)then
c        print*
c        print*,'END OF identify_Ps_for_Exporting_toWest.f'
c        print*,'itime, rank_id ',itime, rank_id
c        print*,'nb_going_East ',nb_going_East
c        print*,'np_going_East ',np_going_East
c        print*,'nb_going_West ',nb_going_West
c        print*,'np_going_West ',np_going_West
c        print*,'np_local ',np_local
c      end if
      
      !stop

      end subroutine
