c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

c
c*********************************************************************
c   !BDR 04/09/06
c   This subroutine performs the following operations:
c
c     (i)   Checks to see if particles are leaving the subdomain of 
c           the processor (i.e. the rank)
c     (ii)  Transfers the particles and all their properties to the
c           adjacent process
c     (iii) On each processor, the ip index system is then reorganised
c           for the beginning of the next timestep
c
c     Note, this subroutine re-organises the index pointer system but does not
c     reorganise the data array

      subroutine index_check_EastWest

      include 'common.2D'

      real(kind=kind_XZ ) xp_temp,zp_temp
      
      double precision dxp_dble, dzp_dble

      integer iBPs_leaving_East(nm_max)  , iBPs_leaving_West(nm_max)
      integer iBPs_entering_East(nm_max) , iBPs_entering_West(nm_max)
      integer iFPs_leaving_East(nm_max)  , iFPs_leaving_West(nm_max)
      integer iFPs_entering_East(nm_max) , iFPs_entering_West(nm_max)

      !Declarations of Particles leaving/entering a processor

      !-- Derived DataTypes --
      real(kind=kind_XZ ) P_xyz_data_leaving_East(nm_max,3)
      real(kind=kind_XZ ) P_xyz_data_leaving_West(nm_max,3)
      real(kind=kind_XZ ) P_xyz_data_entering_East(nm_max,3)
      real(kind=kind_XZ ) P_xyz_data_entering_West(nm_max,3)
      real P_data_leaving_East(nm_max,20)
      real P_data_leaving_West(nm_max,20)
      real P_data_entering_East(nm_max,20)
      real P_data_entering_West(nm_max,20)

      real rhop_sum_data_leaving_East(nm_max)
      real rhop_sum_data_leaving_West(nm_max)
      real rhop_sum_data_entering_East(nm_max)
      real rhop_sum_data_entering_West(nm_max)

      real P_verlet_data_leaving_East(nm_max,4)
      real P_verlet_data_leaving_West(nm_max,4)
      real P_verlet_data_entering_East(nm_max,4)
      real P_verlet_data_entering_West(nm_max,4)
      
      integer i_ABP_leaving_East(nm_max)
      integer i_ABP_leaving_West(nm_max)
      integer i_ABP_entering_East(nm_max)
      integer i_ABP_entering_West(nm_max)

      
c      itime_check = 10e8     !0    !10e8     !
c      i_rank_check = 2

      !if(rank_id.ge.0.and.itime.ge.0)then
      !  print*
      !  print*,'BEGINNING OF index_check.f'
      !  print*,'itime, rank_id ',itime, rank_id
      !  print*
      !end if
      
      !Artificially move the particles from one proc to another
      !if(rank_id.eq.0)then
      ! do ipcount=1,np_local
      !  if(ipcount.le.nb_local)then
      !    i = ip_index_BP_local(ipcount)
      !  else
      !    i = ip_index_FP_local(ipcount-nb_local)
      !  end if
      !    if(xp(i).gt.2.66)then
      !      xp(i) = xp(i) + 0.05
      !    endif
      ! enddo
      !endif


c      call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

     
c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'Index_check'
c        print*,'itime ',itime
c        print*
c        print*,'rank_id ',rank_id
c        print*,'nbf_local        ',nbf_local
c        print*,'nbfm_local  ',nbfm_local
c        print*,'nbfreem_local      ',nbfreem_local
c        print*,'nb_local         ',nb_local
c        print*,'np_local         ',np_local
c        print*
c        !print*,'Cartesian Topology'        
c        nbf_index       = 0
c        nbfm_index = 0
c        nbfreem_index     = 0
c      end if
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(i.eq.-1-rank_id)then
c            write(*,*),'rank_id, east, west ',rank_id, east, west 
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c        
c        call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c      end if
c      !-------------------

      nb_leaving_East = 0
      nb_leaving_West = 0
      np_leaving_East = 0
      np_leaving_West = 0
      nb_entering_East = 0
      nb_entering_West = 0
      np_entering_East = 0
      np_entering_West = 0

      no_of_bfPs_leaving_East = 0
      no_of_bfPs_leaving_West = 0
      no_of_forcedmBPs_leaving_East = 0      !MBP=Forced Moving Boundary Particles
      no_of_forcedmBPs_leaving_West = 0      !MBP=Forced Moving Boundary Particles
      no_of_freemBPs_leaving_East = 0      !MBP=Free Moving Boundary Particles
      no_of_freemBPs_leaving_West = 0      !MBP=Free Moving Boundary Particles
      no_of_bfPs_entering_East = 0
      no_of_bfPs_entering_West = 0
      no_of_forcedmBPs_entering_East = 0     !MBP=Forced Moving Boundary Particles
      no_of_forcedmBPs_entering_West = 0     !MBP=Forced Moving Boundary Particles
      no_of_freemBPs_entering_East = 0     !MBP=Free Moving Boundary Particles
      no_of_freemBPs_entering_West = 0     !MBP=Free Moving Boundary Particles


      no_of_BPs_leaving_East = 0
      no_of_BPs_leaving_West = 0
      no_of_BPs_entering_East = 0
      no_of_BPs_entering_West = 0

      no_of_FPs_leaving_East = 0
      no_of_FPs_leaving_West = 0
      no_of_FPs_entering_East = 0
      no_of_FPs_entering_West = 0

      do i=1,nm_max
        iBPs_leaving_East(i) = 0
        iBPs_leaving_West(i) = 0
        iFPs_leaving_East(i) = 0
        iFPs_leaving_West(i) = 0

        iBPs_entering_East(i) = 0
        iBPs_entering_West(i) = 0
        iFPs_entering_East(i) = 0
        iFPs_entering_West(i) = 0
      end do
      
      if(i_densityFilter.gt.0.and.
     &        itime/ndt_FilterPerform*ndt_FilterPerform.eq.itime)then   !ndt_FilterPerform set in getdata.f 
        i_filterTransfer_Perform = 1
      else
        i_filterTransfer_Perform = 0
      endif
      
      if(i_MPI_Adapt.eq.1.and.n_procs_x.gt.1)then
        if(itime/n_itime_Adapt*n_itime_Adapt.eq.itime
     &                      .or.itime.eq.itime_init)then
          call Adapt_Check_EastWest
        endif
      endif
c
c      itop_free_index_temp = itop_free_index  ! BDR itop_free_index
c                                             ! points to the last member in list of free
c                                             ! array locations (ifree_index)

      !- Set xmax_rank_check to account for periodicity - xmax_container_double
      if((iCartTop_periodicity(1).or.i_periodicOBs(1).eq.1).and.
     &   rank_idx.eq.n_procs_x-1)then
        xmax_rank_check = xmax_container  !_double
      else
        xmax_rank_check = xmax_rank
      endif

c     ***  check to see if particles leave the domain  ***
      !- Boundary Particles -
      !- Fixed Boundary Particles -
      do ipcount = 1,nbf_local
        i=ip_index_BP_local_1(ipcount)
        if(xp(i).gt.xmax_rank_check)then  !Check East boundary
          no_of_BPs_leaving_East = no_of_BPs_leaving_East + 1
          iBPs_leaving_East(no_of_BPs_leaving_East) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_bfPs_leaving_East = no_of_bfPs_leaving_East +1
        else if(xp(i).lt.xmin_rank)then  !Check West boundary
          no_of_BPs_leaving_West = no_of_BPs_leaving_West + 1
          iBPs_leaving_West(no_of_BPs_leaving_West) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_bfPs_leaving_West = no_of_bfPs_leaving_West +1
        end if
      end do
      !- Forced Boundary Particles -
      do ipcount = nbf_local + 1,nbfm_local
        i=ip_index_BP_local_2(ipcount - nbf_local)
        if(xp(i).gt.xmax_rank_check)then  !Check East boundary
          no_of_BPs_leaving_East = no_of_BPs_leaving_East + 1
          iBPs_leaving_East(no_of_BPs_leaving_East) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_forcedmBPs_leaving_East = 
     &     no_of_forcedmBPs_leaving_East+1
        else if(xp(i).lt.xmin_rank)then  !Check West boundary
          no_of_BPs_leaving_West = no_of_BPs_leaving_West + 1
          iBPs_leaving_West(no_of_BPs_leaving_West) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_forcedmBPs_leaving_West = 
     &     no_of_forcedmBPs_leaving_West+1
        end if
      end do
      !- Free-Moving Boundary Particles -
      do ipcount = nbfm_local + 1,nb_local
        i=ip_index_BP_local_3(ipcount - nbfm_local)
        if(xp(i).gt.xmax_rank_check)then  !Check East boundary
          no_of_BPs_leaving_East = no_of_BPs_leaving_East + 1
          iBPs_leaving_East(no_of_BPs_leaving_East) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_freemBPs_leaving_East   = no_of_freemBPs_leaving_East +1
        else if(xp(i).lt.xmin_rank)then  !Check West boundary
          no_of_BPs_leaving_West = no_of_BPs_leaving_West + 1
          iBPs_leaving_West(no_of_BPs_leaving_West) = i
          iBPs_top_free_index = iBPs_top_free_index +1
          iBPs_free_index(iBPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
          no_of_freemBPs_leaving_West   = no_of_freemBPs_leaving_West +1
        end if
      end do    
      !- Fluid Particles -
      do ipcount = nb_local + 1,np_local
        i=ip_index_FP_local(ipcount - nb_local)
        if(xp(i).gt.xmax_rank_check)then  !Check East boundary
          no_of_FPs_leaving_East = no_of_FPs_leaving_East + 1
          iFPs_leaving_East(no_of_FPs_leaving_East) = i
          iFPs_top_free_index = iFPs_top_free_index +1
          iFPs_free_index(iFPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
        else if(xp(i).lt.xmin_rank)then  !Check West boundary
          no_of_FPs_leaving_West = no_of_FPs_leaving_West + 1
          iFPs_leaving_West(no_of_FPs_leaving_West) = i
          iFPs_top_free_index = iFPs_top_free_index +1
          iFPs_free_index(iFPs_top_free_index) = i
          iarray_loc_in_use(i) = 0
        end if

      end do

      nb_leaving_East = no_of_BPs_leaving_East
      np_leaving_East = no_of_BPs_leaving_East +no_of_FPs_leaving_East
      nb_leaving_West = no_of_BPs_leaving_West
      np_leaving_West = no_of_BPs_leaving_West +no_of_FPs_leaving_West

c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then
c            print*
c            print*,'itime ',itime
c            write(*,*),'rank_id ', rank_id
c            write(*,*),'nb_leaving_East, nb_leaving_West ',
c     &                  nb_leaving_East, nb_leaving_West 
c            write(*,*),'np_leaving_East, np_leaving_West ',
c     &                  np_leaving_East, np_leaving_West 
c            print*,'no_of_bfPs_leaving_East ',no_of_bfPs_leaving_East
c            print*,'no_of_bfPs_leaving_West ',no_of_bfPs_leaving_West
c            print*,'no_of_forcedmBPs_leaving_East ',
c     &              no_of_forcedmBPs_leaving_East
c            print*,'no_of_forcedmBPs_leaving_West ',
c     &              no_of_forcedmBPs_leaving_West
c            print*,'no_of_freemBPs_leaving_East ',no_of_freemBPs_leaving_East
c            print*,'no_of_freemBPs_leaving_West ',no_of_freemBPs_leaving_West
c            print*,'no_of_FPs_leaving_East ',no_of_FPs_leaving_East
c            print*,'no_of_FPs_leaving_West ',no_of_FPs_leaving_West
c            print*,'nbf_local        ',nbf_local
c            print*,'nbfm_local  ',nbfm_local
c            print*,'nbfreem_local      ',nbfreem_local
c            print*,'nb_local         ',nb_local
c            print*,'np_local         ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if
c      !----------------------

      if(nb_leaving_East.gt.nm_max.or.nb_leaving_West.gt.nm_max.or.
     &   np_leaving_East.gt.nm_max.or.np_leaving_West.gt.nm_max.or.
     &   nb_leaving_East.lt.0.or.nb_leaving_West.lt.0.or.
     &   np_leaving_East.lt.0.or.np_leaving_West.lt.0)then
        print*
        print*,'Index_check.f'      
        print*,'n#_leaving_##st.gt.nm_max OR'
        print*,'n#_leaving_##st.lt.0'
        print*,'itime ',itime
        write(*,*),'rank_id ', rank_id
        print*,'mn_max ',nm_max
        write(*,*),'nb_leaving_East, nb_leaving_West ',
     &              nb_leaving_East, nb_leaving_West 
        write(*,*),'np_leaving_East, np_leaving_West ',
     &              np_leaving_East, np_leaving_West 
        print*,'no_of_bfPs_leaving_East ',no_of_bfPs_leaving_East
        print*,'no_of_bfPs_leaving_West ',no_of_bfPs_leaving_West
        print*,'no_of_forcedmBPs_leaving_East ',
     &          no_of_forcedmBPs_leaving_East
        print*,'no_of_forcedmBPs_leaving_West ',
     &          no_of_forcedmBPs_leaving_West
        print*,'no_of_freemBPs_leaving_East',
     &          no_of_freemBPs_leaving_East
        print*,'no_of_freemBPs_leaving_West',
     &          no_of_freemBPs_leaving_West
        print*,'no_of_FPs_leaving_East ',no_of_FPs_leaving_East
        print*,'no_of_FPs_leaving_West ',no_of_FPs_leaving_West
        print*,'nbf_local        ',nbf_local
        print*,'nbfm_local  ',nbfm_local
        print*,'nbfreem_local      ',nbfreem_local
        print*,'nb_local         ',nb_local
        print*,'np_local         ',np_local
        stop
      end if

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      !--- PACK DATA TO BE SENT EAST ---
c      !-- Debug Code------
c      if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c        print*
c        print*,'Packing Particle Data leaving East'
c        print*,'rank_id ',rank_id
c        print*,'nb_leaving_East ',nb_leaving_East
c        print*,'np_leaving_East ',np_leaving_East
c        print*,'no_of_bfPs_leaving_East      ',no_of_bfPs_leaving_East
c        print*,'no_of_forcedmBPs_leaving_East ',
c     &          no_of_forcedmBPs_leaving_East
c        print*,'no_of_freemBPs_leaving_East      ',no_of_freemBPs_leaving_East
c        print*,'no_of_BPs_leaving_East ',no_of_BPs_leaving_East
c        print*,'no_of_FPs_leaving_East ',no_of_FPs_leaving_East
c        print*,' ii,   i, xp_leaving_East(ii), zp_leaving_East(ii) '
c      end if
      do ii=1,np_leaving_East
        if(ii.le.nb_leaving_East)then        
          i = iBPs_leaving_East(ii)
        else
          i = iFPs_leaving_East(ii-nb_leaving_East)
        end if        
        P_xyz_data_leaving_East(ii,1)  = xp(i)
c                                      = yp(i)
        P_xyz_data_leaving_East(ii,2)  = zp(i)        
        P_data_leaving_East(ii,1)  = up(i)
c                                  = vp(i)
        P_data_leaving_East(ii,2)  = wp(i)        
        P_data_leaving_East(ii,3)  = rhop(i)
        P_data_leaving_East(ii,4)  = p(i)        
        P_data_leaving_East(ii,5)  = pm(i)        
        P_data_leaving_East(ii,6)  = pVol(i)        
        P_data_leaving_East(ii,7)  = cs(i)        
        P_data_leaving_East(ii,8)  = sum_wab(i)
        P_data_leaving_East(ii,9)  = tau_xx_uno_rho2(i)
c                                  = tau_xy_uno_rho2(i)
        P_data_leaving_East(ii,10) = tau_xz_uno_rho2(i)
c                                  = tau_yy_uno_rho2(i)
c                                  = tau_yz_uno_rho2(i)
        P_data_leaving_East(ii,11) = tau_zz_uno_rho2(i)
        P_data_leaving_East(ii,12) = TEp(i)
        P_data_leaving_East(ii,13) = dudx_CSPH(i)
c                                  = dudy_CSPH(i)
        P_data_leaving_East(ii,14) = dudz_CSPH(i)
c                                  = dvdx_CSPH(i)
c                                  = dvdy_CSPH(i)
c                                  = dvdz_CSPH(i)
        P_data_leaving_East(ii,15) = dwdx_CSPH(i)
c                                  = dwdy_CSPH(i)
        P_data_leaving_East(ii,16) = dwdz_CSPH(i)
        P_data_leaving_East(ii,17) = drhodx_CSPH(i)
c                                  = drhody_CSPH(i)
        P_data_leaving_East(ii,18) = drhodz_CSPH(i)
        P_data_leaving_East(ii,19) = dTEdx_CSPH(i)
c                                  = dTEdy_CSPH(i)
        P_data_leaving_East(ii,20) = dTEdz_CSPH(i)
        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum_data_leaving_East(ii) = rhop_sum(i)
        endif
        !Updating Normals Data
        if(ii.le.nb_leaving_East)then
          i_ABP_leaving_East(ii) = iabsolute_BP_Pointer_Index(i)    
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          P_verlet_data_leaving_East(ii,1) = uo(i)
c                                          = vo(i)
          P_verlet_data_leaving_East(ii,2) = wo(i)
          P_verlet_data_leaving_East(ii,3) = rhoo(i)
          P_verlet_data_leaving_East(ii,4) = TEo(i)
        endif

c        !-- Debug Code------
c        if(rank_id.eq.2.and.itime.ge.itime_check)then
c          print*,'ii, i',ii, i
c          print*,'xp,zp_leaving_East(ii) ',
c     &    P_xyz_Data_leaving_East(ii,1),
c     &    P_xyz_Data_leaving_East(ii,2),
c
c          print*,'i_particleType(i) ',i_particleType(i)
c          if(ii.le.nb_leaving_East)then
c            print*,'iabsolute_BP_Pointer_Index(i) ',
c     &              i_ABP_leaving_East(ii)
c          end if
c        end if
c        !-------------------

      end do

      !--- PACK DATA TO BE SENT WEST ---
c      !-- Debug Code------
c      if(rank_id.eq.3.and.itime.ge.itime_check)then
c        print*
c        print*,'Packing Particle Data leaving West'
c        print*,'rank_id ',rank_id
c        print*,'nb_leaving_West ',nb_leaving_West
c        print*,'np_leaving_West ',np_leaving_West
c        print*,'no_of_bfPs_leaving_West      ',no_of_bfPs_leaving_West
c        print*,'no_of_forcedmBPs_leaving_West ',
c     &          no_of_forcedmBPs_leaving_West
c        print*,'no_of_freemBPs_leaving_West      ',no_of_freemBPs_leaving_West
c        print*,'no_of_BPs_leaving_West ',no_of_BPs_leaving_West
c        print*,'no_of_FPs_leaving_West ',no_of_FPs_leaving_West
c        print*,' ii,   i, xp_leaving_West(ii), zp_leaving_West(ii) '
c      end if
c      !--------------------
      do ii=1,np_leaving_West
        if(ii.le.nb_leaving_West)then        
          i = iBPs_leaving_West(ii)
        else
          i = iFPs_leaving_West(ii-nb_leaving_West)
        end if        
        P_xyz_data_leaving_West(ii,1)  = xp(i)
c                                      = yp(i)
        P_xyz_data_leaving_West(ii,2)  = zp(i)        
        P_data_leaving_West(ii,1)  = up(i)
c                                  = vp(i)
        P_data_leaving_West(ii,2)  = wp(i)        
        P_data_leaving_West(ii,3)  = rhop(i)
        P_data_leaving_West(ii,4)  = p(i)        
        P_data_leaving_West(ii,5)  = pm(i)        
        P_data_leaving_West(ii,6)  = pVol(i)        
        P_data_leaving_West(ii,7)  = cs(i)        
        P_data_leaving_West(ii,8)  = sum_wab(i)
        P_data_leaving_West(ii,9)  = tau_xx_uno_rho2(i)
c                                  = tau_xy_uno_rho2(i)
        P_data_leaving_West(ii,10) = tau_xz_uno_rho2(i)
c                                  = tau_yy_uno_rho2(i)
c                                  = tau_yz_uno_rho2(i)
        P_data_leaving_West(ii,11) = tau_zz_uno_rho2(i)
        P_data_leaving_West(ii,12) = TEp(i)
        P_data_leaving_West(ii,13) = dudx_CSPH(i)
c                                   = dudy_CSPH(i)
        P_data_leaving_West(ii,14) = dudz_CSPH(i)
c                                   = dvdx_CSPH(i)
c                                   = dvdy_CSPH(i)
c                                   = dvdz_CSPH(i)
        P_data_leaving_West(ii,15) = dwdx_CSPH(i)
c                                   = dwdy_CSPH(i)
        P_data_leaving_West(ii,16) = dwdz_CSPH(i)
        P_data_leaving_West(ii,17) = drhodx_CSPH(i)
c                                   = drhody_CSPH(i)
        P_data_leaving_West(ii,18) = drhodz_CSPH(i)
        P_data_leaving_West(ii,19) = dTEdx_CSPH(i)
c                                  = dTEdy_CSPH(i)
        P_data_leaving_West(ii,20) = dTEdz_CSPH(i)
        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum_data_leaving_West(ii) = rhop_sum(i)
        endif
        !Updating Normals Data
        if(ii.le.nb_leaving_West)then
          i_ABP_leaving_West(ii) = iabsolute_BP_Pointer_Index(i)    
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          P_verlet_data_leaving_West(ii,1) = uo(i)
c                                          = vo(i)
          P_verlet_data_leaving_West(ii,2) = wo(i)
          P_verlet_data_leaving_West(ii,3) = rhoo(i)
          P_verlet_data_leaving_West(ii,4) = TEo(i)
        endif

c        !-- Debug Code------
c        if(rank_id.eq.2.and.itime.ge.itime_check)then   !if(rank_id.eq.-10)then
c          print*,'ii, i',ii, i
c          print*,'xp_leaving_West(ii), zp_leaving_West(ii) ',
c     &    P_xyz_Data_leaving_West(ii,1), 
c     &    P_xyz_Data_leaving_West(ii,2) 
c
c          print*,'i_particleType(i) ',i_particleType(i)
c          if(ii.le.nb_leaving_West)then
c            print*,'iabsolute_BP_Pointer_Index(i) ',
c     &              i_ABP_leaving_West(ii)
c          end if
c        end if
c
c        if(rank_id.eq.2.and.itime.ge.itime_check)then   !if(rank_id.eq.-10)then
c          if(i.eq.98)then
c            print*,'particle i= ',i,' leaving rank_id ',rank_id
c            print*,'ii leaving east ',ii
c            print*,'P_xyz_data_leaving_West(ii,1)  = xp(i)          ',
c     &              P_xyz_data_leaving_West(ii,1)
c            print*,'P_xyz_data_leaving_West(ii, )  = yp(i)          ',
c     &              P_xyz_data_leaving_West(ii, )
c            print*,'P_xyz_data_leaving_West(ii,2)  = zp(i)          ',
c     &              P_xyz_data_leaving_West(ii,2)
c            print*,'P_data_leaving_West(ii,1)  = up(i)              ',
c     &              P_data_leaving_West(ii,1)
c            print*,'P_data_leaving_West(ii, )  = vp(i)              ',
c     &              P_data_leaving_West(ii, )
c            print*,'P_data_leaving_West(ii,2)  = wp(i)              ',
c     &              P_data_leaving_West(ii,2)
c            print*,'P_data_leaving_West(ii,3)  = rhop(i)            ',
c     &              P_data_leaving_West(ii,3)
c            print*,'P_data_leaving_West(ii,4)  = p(i)               ',
c     &              P_data_leaving_West(ii,4)
c            print*,'P_data_leaving_West(ii,5)  = pm(i)              ',
c     &              P_data_leaving_West(ii,5)
c            print*,'P_data_leaving_West(ii,6)  = pVol(i)            ',
c     &              P_data_leaving_West(ii,6)
c            print*,'P_data_leaving_West(ii,7)  = cs(i)              ',
c     &              P_data_leaving_West(ii,7)
c            print*,'P_data_leaving_West(ii,8) = sum_wab(i)          ',
c     &              P_data_leaving_West(ii,8)
c            print*,'P_data_leaving_West(ii,9)  = tau_xx_uno_rho2(i) ',
c     &              P_data_leaving_West(ii,9)
c            print*,'P_data_leaving_West(ii, )  = tau_xy_uno_rho2(i) ',
c     &              P_data_leaving_West(ii, )
c            print*,'P_data_leaving_West(ii,10) = tau_xz_uno_rho2(i) ',
c     &              P_data_leaving_West(ii,10)
c            print*,'P_data_leaving_West(ii,  ) = tau_yy_uno_rho2(i) ',
c     &              P_data_leaving_West(ii,  )
c            print*,'P_data_leaving_West(ii,  ) = tau_yz_uno_rho2(i) ',
c     &              P_data_leaving_West(ii,  )
c            print*,'P_data_leaving_West(ii,11) = tau_zz_uno_rho2(i) ',
c     &              P_data_leaving_West(ii,11)
c            print*,'dudx_CSPH(i) ',dudx_CSPH(i)
cc           print*,'dudy_CSPH(i) ',dudy_CSPH(i)
c            print*,'dudz_CSPH(i) ',dudz_CSPH(i)
cc           print*,'dvdx_CSPH(i) ',dvdx_CSPH(i)
cc           print*,'dvdy_CSPH(i) ',dvdy_CSPH(i)
cc           print*,'dvdz_CSPH(i) ',dvdz_CSPH(i)
c            print*,'dwdx_CSPH(i) ',dwdx_CSPH(i)
cc           print*,'dwdy_CSPH(i) ',dwdy_CSPH(i)
c            print*,'dwdz_CSPH(i) ',dwdz_CSPH(i)
c            print*,'udot(i) ',udot(i)
cc           print*,'vdot(i) ',vdot(i)
c            print*,'wdot(i) ',wdot(i)
c            print*,'rdot(i) ',rdot(i)
c            !stop
c          endif
c        endif
c        !------------------------
      
      end do
      !--------------------------------------------------

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      !--- Transfer Particles to neighbouring Processes ---

      !-- SendRecv particle numbers --
      !Send to East neighbour
      call MPI_SENDRECV(
     &  nb_leaving_East  , 1, MPI_INTEGER, east, 1,
     &  nb_entering_West , 1, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  np_leaving_East  , 1, MPI_INTEGER, east, 1,
     &  np_entering_West , 1, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_bfPs_leaving_East  , 1, MPI_INTEGER, east, 1,
     &  no_of_bfPs_entering_West , 1, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_forcedmBPs_leaving_East  , 1, MPI_INTEGER, east, 1,
     &  no_of_forcedmBPs_entering_West , 1, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_freemBPs_leaving_East  , 1, MPI_INTEGER, east, 1,
     &  no_of_freemBPs_entering_West , 1, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      !Send to West Neighbour
      call MPI_SENDRECV(
     &  nb_leaving_West  , 1, MPI_INTEGER, west, 0,
     &  nb_entering_East , 1, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  np_leaving_West  , 1, MPI_INTEGER, west, 0,
     &  np_entering_East , 1, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_bfPs_leaving_West  , 1, MPI_INTEGER, west, 0,
     &  no_of_bfPs_entering_East , 1, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_forcedmBPs_leaving_West  , 1, MPI_INTEGER, west, 0,
     &  no_of_forcedmBPs_entering_East , 1, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  no_of_freemBPs_leaving_West  , 1, MPI_INTEGER, west, 0,
     &  no_of_freemBPs_entering_East , 1, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

      !-- SendRecv Particle Data --
      !Send to East neighbour
      call MPI_SENDRECV(
     &  i_ABP_leaving_East  , nb_leaving_East , MPI_INTEGER, east, 1,
     &  i_ABP_entering_West , nb_entering_West, MPI_INTEGER, west, 1,
     &  grid_comm, status, ierr)
      !---Combined SendRecv's to East Neighbour ---
      call MPI_SENDRECV(
     &  P_xyz_data_leaving_East     , 1 , iP_xyzDataTypeB, east, 240,
     &  P_xyz_data_entering_West    , 1 , iP_xyzDataTypeB, west, 240,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  P_data_leaving_East     , 1 , iP_DataTypeIC , east, 241,
     &  P_data_entering_West    , 1 , iP_DataTypeIC , west, 241,
     &  grid_comm, status, ierr)
      if(i_filterTransfer_Perform.eq.1)then
       call MPI_SENDRECV(
     &  rhop_sum_data_leaving_East , np_leaving_East , MPI_REAL, east,1,
     &  rhop_sum_data_entering_West, np_entering_West, MPI_REAL, west,1,
     &  grid_comm, status, ierr)
      endif
      !- To be used with Verlet algorithm -
      if(i_algorithm.eq.2) then 
        call MPI_SENDRECV(
     &   P_verlet_data_leaving_East     , 1 , iP_DataTypeV, east, 242,
     &   P_verlet_data_entering_West    , 1 , iP_DataTypeV, west, 242,
     &   grid_comm, status, ierr)
      endif

      !end if   !End of:   if(np_leaving_east.gt.0.or.np_entering_West.gt.0)then

      !Send to West Neighbour
      call MPI_SENDRECV(
     &  i_ABP_leaving_West  , nb_leaving_West , MPI_INTEGER, west, 0,
     &  i_ABP_entering_East , nb_entering_East, MPI_INTEGER, east, 0,
     &  grid_comm, status, ierr)
      !---Combined SendRecv's to West Neighbour ---
      call MPI_SENDRECV(
     &  P_xyz_data_leaving_West     , 1 , iP_xyzDataTypeB, west, 243,
     &  P_xyz_data_entering_East    , 1 , iP_xyzDataTypeB, east, 243,
     &  grid_comm, status, ierr)
      call MPI_SENDRECV(
     &  P_data_leaving_West     , 1 , iP_DataTypeIC , west, 244,
     &  P_data_entering_East    , 1 , iP_DataTypeIC , east, 244,
     &  grid_comm, status, ierr)
      if(i_filterTransfer_Perform.eq.1)then
       call MPI_SENDRECV(
     &  rhop_sum_data_leaving_West , np_leaving_West , MPI_REAL, west,0,
     &  rhop_sum_data_entering_East, np_entering_East, MPI_REAL, east,0,
     &  grid_comm, status, ierr)
      endif
      !- To be used with Verlet algorithm -
      if(i_algorithm.eq.2) then 
        call MPI_SENDRECV(
     &   P_verlet_data_leaving_West     , 1 , iP_DataTypeV, west, 245,
     &   P_verlet_data_entering_East    , 1 , iP_DataTypeV, east, 245,
     &   grid_comm, status, ierr)
      endif
      !end if   !End of:  if(np_leaving_West.gt.0.or.np_entering_East.gt.0)then
c     -----------------------------------------

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

c     --- Prepare Data arrays for Unpacking ---
      net_nb_change = nb_entering_East + nb_entering_West
     &              - nb_leaving_East  - nb_leaving_West
      net_np_change = np_entering_East + np_entering_West
     &              - np_leaving_East  - np_leaving_West
      net_nbf_change = 
     &     no_of_bfPs_entering_East + no_of_bfPs_entering_West
     &   - no_of_bfPs_leaving_East  - no_of_bfPs_leaving_West
      net_nbfm_change =  
     &   no_of_forcedmBPs_entering_East + no_of_forcedmBPs_entering_West
     & - no_of_forcedmBPs_leaving_East  - no_of_forcedmBPs_leaving_West
      net_freemBP_change =  
     &   no_of_freemBPs_entering_East + no_of_freemBPs_entering_West
     & - no_of_freemBPs_leaving_East  - no_of_freemBPs_leaving_West

      !nptemp = np_local - no_of_Ps_leaving
      nbtemp = nb_local - nb_leaving_East - nb_leaving_West
      nptemp = np_local - np_leaving_East - np_leaving_West

      nbf_temp = nbf_local
     &  - no_of_bfPs_leaving_East       - no_of_bfPs_leaving_West
      nbfm_temp = nbfm_local
     &  - no_of_forcedmBPs_leaving_East - no_of_forcedmBPs_leaving_West
      nbfreem_temp = nbfreem_local
     &  - no_of_freemBPs_leaving_East   - no_of_freemBPs_leaving_West
      nFP_temp = np_local - nb_local
     & - no_of_FPs_leaving_East        - no_of_FPs_leaving_West
      
      nbf_local     = nbf_local  + net_nbf_change
      nbfm_local    = nbfm_local + net_nbf_change
     &                           + net_nbfm_change
      nbfreem_local = nbfreem_local + net_nbf_change
     &                              + net_nbfm_change
     &                              + net_freemBP_change 
c      nb_local        = nb_local  + net_nbf_change
c     &                            + net_nbfm_change
c     &                            + net_freemBP_change
c      nbfp1_local = nbf_local + 1
c      nbp1_local  = nb_local  + 1
      
      if(net_nbf_change.ne.0.and.i_MPI_Adapt.ne.1)then
        print*
        print*,'Error in index_check.f'
        print*,'net_nbf_change.ne.0'
        print*,'itime ',itime
        print*,'rank_id        ',rank_id
        print*,'nbf_local      ',nbf_local
        print*,'nb_local       ',nb_local
        print*,'np_local       ',np_local
        print*,'net_nbf_change ',net_nbf_change
        stop
      endif

      nb_local_temp = nb_local + net_nb_change
      np_local_temp = np_local + net_np_change
      
c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then
c            print*
c            print*,'rank_id ',rank_id
c            print*,'BPs leaving & entering'
c            print*,'nb_leaving_East  ',nb_leaving_East
c            print*,'nb_leaving_West  ',nb_leaving_West
c            print*,'nb_entering_East ',nb_entering_East
c            print*,'nb_entering_West ',nb_entering_West
c            print*,'Ps leaving & entering'
c            print*,'np_leaving_East  ',np_leaving_East
c            print*,'np_leaving_West  ',np_leaving_West
c            print*,'np_entering_East ',np_entering_East
c            print*,'np_entering_West ',np_entering_West
c            print*,'nbtemp = nb_local - nb_leaving ',nbtemp
c            print*,'nptemp = np_local - np_leaving ',nptemp
c            print*,'net_nb_change ',net_nb_change
c            print*,'net_np_change ',net_np_change
c            print*,'nb_local_temp ',nb_local_temp
c            print*,'net_nbf_change        ',net_nbf_change
c            print*,'net_nbfm_change  ',net_nbfm_change
c            print*,'net_freemBP_change       ',net_freemBP_change
c            print*,'nbf_local        ',nbf_local
c            print*,'nbfm_local  ',nbfm_local
c            print*,'nbfreem_local      ',nbfreem_local
c            print*,'nb_local         ',nb_local
c            print*,'np_local         ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if

c     ***  fill up free spaces with new particles  ***

c     --- UNPACK DATA ENTERING WEST ---
c      !-- Debug Code------
c      if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c        print*
c        print*,'UnPacking Particle Data entering West'
c        print*,'rank_id ',rank_id
c        print*,'nb_entering_West ',nb_entering_West
c        print*,'np_entering_West ',np_entering_West
c        print*,'nb_local_temp    ',nb_local_temp
c        print*,'nbf_local          ',nbf_local
c        print*,'nbfm_local    ',nbfm_local
c        print*,'nbfreem_local        ',nbfreem_local
c        print*,'nbtemp          ',nbtemp
c        print*,'nptemp          ',nptemp
c        print*,'nbf_temp        ',nbf_temp
c        print*,'nbfm_temp  ',nbfm_temp
c        print*,'nbfreem_temp      ',nbfreem_temp
c      end if
c      !------------------------

      nbf_count       = 0
      nbfm_count      = 0
      nbfreem_count   = 0
      nFP_count       = 0
      
      do i = 1,np_entering_West

        nptemp = nptemp +1

        if(i.le.nb_entering_West)then
          if(iBPs_top_free_index .gt. 0)then
            ifree_index_ptr = iBPs_free_index(iBPs_top_free_index)
            iBPs_top_free_index = iBPs_top_free_index -1
          else if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nbtemp = nbtemp +1
          ip_index_BP_local(nbtemp) = ifree_index_ptr
          !- Indexing -
          if(nbf_count.lt.no_of_bfPs_entering_West)then   !if(nbf_temp.lt.nbf_local)then
            nbf_count = nbf_count + 1   !nbf_temp = nbf_temp + 1
            nbf_index = nbf_count + nbf_temp
            i_particleType(ifree_index_ptr) = 1   !Fixed Boundary Particle
            ip_index_BP_local_1(nbf_index) = ifree_index_ptr
          else if(nbfm_count.lt.no_of_forcedmBPs_entering_West)then
            nbfm_count = nbfm_count + 1
            nbfm_index = nbfm_count + nbfm_temp
            i_particleType(ifree_index_ptr) = 2   !Forced Moving Boundary Particle
            ip_index_BP_local_2(nbfm_index) = ifree_index_ptr
          else if(nbfreem_count.lt.no_of_freemBPs_entering_West)then
            nbfreem_count = nbfreem_count + 1
            nbfreem_index = nbfreem_count + nbfreem_temp
            i_particleType(ifree_index_ptr) = 3   !Free Moving Boundary Particle
            ip_index_BP_local_3(nbfreem_index) = ifree_index_ptr
          end if
        else
          if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nFP_count = nFP_count + 1
          nFP_index = nFP_count + nFP_temp
          ip_index_FP_local(nFP_index) = ifree_index_ptr
          !ip_index_FP_local(nptemp-nb_local_temp) = ifree_index_ptr
          i_particleType(ifree_index_ptr) = 4   !Fluid Particle
        end if

        if(iarray_loc_in_use(ifree_index_ptr).ne.0)then
          print*
          print*,'ERROR in Index_check.f: particles entering West'
          print*,'iarray_loc_in_use(ifree_index_ptr).ne.0'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'nb_local ',nb_local
          print*,'nbtemp  ',nbtemp
          print*,'nptemp  ',nptemp
          print*,'iBPs_top_free_index ',iBPs_top_free_index
          print*,'iFPs_top_free_index ',iFPs_top_free_index
          print*,'ifree_index_ptr ',ifree_index_ptr
          print*,'iarray_loc_in_use(ifree_index_ptr)',
     &            iarray_loc_in_use(ifree_index_ptr)
          print*,'BPs leaving & entering'
          print*,'nb_leaving_East  ',nb_leaving_East
          print*,'nb_leaving_West  ',nb_leaving_West
          print*,'nb_entering_East ',nb_entering_East
          print*,'nb_entering_West ',nb_entering_West
          print*,'Ps leaving & entering'
          print*,'np_leaving_East  ',np_leaving_East
          print*,'np_leaving_West  ',np_leaving_West
          print*,'np_entering_East ',np_entering_East
          print*,'np_entering_West ',np_entering_West
          print*,'nbtemp = nb_local - nb_leaving ',nbtemp
          print*,'nptemp = np_local - np_leaving ',nptemp
          print*,'net_nb_change ',net_nb_change
          print*,'net_np_change ',net_np_change
          print*,'nb_local_temp ',nb_local_temp
          print*,'net_nbf_change        ',net_nbf_change
          print*,'net_nbfm_change  ',net_nbfm_change
          print*,'net_freemBP_change       ',net_freemBP_change
          print*,'nbf_local        ',nbf_local
          print*,'nbfm_local  ',nbf_local
          print*,'nbfreem_local      ',nbf_local
          print*,'nb_local         ',nb_local
          stop     
        end if
        
        xp(ifree_index_ptr)              = P_xyz_data_entering_West(i,1)
c       yp(ifree_index_ptr)              =
        zp(ifree_index_ptr)              = P_xyz_data_entering_West(i,2)
        up(ifree_index_ptr)              = P_data_entering_West(i,1)
c       vp(ifree_index_ptr)              =
        wp(ifree_index_ptr)              = P_data_entering_West(i,2)
        rhop(ifree_index_ptr)            = P_data_entering_West(i,3)
        p(ifree_index_ptr)               = P_data_entering_West(i,4)
        pm(ifree_index_ptr)              = P_data_entering_West(i,5)
        pVol(ifree_index_ptr)            = P_data_entering_West(i,6)
        cs(ifree_index_ptr)              = P_data_entering_West(i,7)
        sum_wab(ifree_index_ptr)         = P_data_entering_West(i,8)
        tau_xx_uno_rho2(ifree_index_ptr) = P_data_entering_West(i,9) 
c       tau_xy_uno_rho2(ifree_index_ptr) =
        tau_xz_uno_rho2(ifree_index_ptr) = P_data_entering_West(i,10)
c       tau_yy_uno_rho2(ifree_index_ptr) =
c       tau_yz_uno_rho2(ifree_index_ptr) =
        tau_zz_uno_rho2(ifree_index_ptr) = P_data_entering_West(i,11)
        TEp(ifree_index_ptr)             = P_data_entering_West(i,12)
        dudx_CSPH(ifree_index_ptr)       = P_data_entering_West(i,13)
c       dudy_CSPH(ifree_index_ptr)       = 
        dudz_CSPH(ifree_index_ptr)       = P_data_entering_West(i,14)
c       dvdx_CSPH(ifree_index_ptr)       = 
c       dvdy_CSPH(ifree_index_ptr)       = 
c       dvdz_CSPH(ifree_index_ptr)       = 
        dwdx_CSPH(ifree_index_ptr)       = P_data_entering_West(i,15)
c       dwdy_CSPH(ifree_index_ptr)       = 
        dwdz_CSPH(ifree_index_ptr)       = P_data_entering_West(i,16)
        drhodx_CSPH(ifree_index_ptr)     = P_data_entering_West(i,17)
c       drhody_CSPH(ifree_index_ptr)     = 
        drhodz_CSPH(ifree_index_ptr)     = P_data_entering_West(i,18)
        dTEdx_CSPH(ifree_index_ptr)      = P_data_entering_West(i,19)
c       dTEdy_CSPH(ifree_index_ptr)      = 
        dTEdz_CSPH(ifree_index_ptr)      = P_data_entering_West(i,20)
        iarray_loc_in_use(ifree_index_ptr) = 1
        iflag(ifree_index_ptr) = 1

        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum(ifree_index_ptr) = rhop_sum_data_entering_West(i)
        endif
        if(i.le.nb_entering_West)then
          iabsolute_BP_Pointer_Index(ifree_index_ptr) = 
     &                           i_ABP_entering_West(i)
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          uo(ifree_index_ptr)   = P_verlet_data_entering_West(i,1)
c         vo(ifree_index_ptr)   =
          wo(ifree_index_ptr)   = P_verlet_data_entering_West(i,2)
          rhoo(ifree_index_ptr) = P_verlet_data_entering_West(i,3)
          TEo(ifree_index_ptr)  = P_verlet_data_entering_West(i,4)
        endif

        !-- MPI Periodic OB correction --
        !-- Periodic OB correction X-Direction --
        if(iCartTop_periodicity(1))then
         if(rank_idx.eq.0.and.xp(ifree_index_ptr).gt.xmax_container)then
           dxp_dble = dble(xp(ifree_index_ptr)) - xmax_container_double    
           xp(ifree_index_ptr) = real(dxp_dble +xmin_container_double)
         else if(rank_idx.eq.n_procs_x-1.and.
     &                       xp(ifree_index_ptr).lt.xmin_container)then
           dxp_dble = xmin_container_double - dble(xp(ifree_index_ptr))   
           xp(ifree_index_ptr) = real(xmax_container_double - dxp_dble)
         end if
        end if
!         !-- Periodic OB correction Y-Direction --
!         if(iCartTop_periodicity(2))then
!          if(rank_idy.eq.0.and.yp(ifree_index_ptr).gt.ymax_container)then
!            dyp_dble = dble(yp(ifree_index_ptr)) - ymax_container_double    
!            yp(ifree_index_ptr) = real(dyp_dble +ymin_container_double)
!          else if(rank_idy.eq.n_procs_y-1.and.
!      &                       yp(ifree_index_ptr).lt.ymin_container)then
!            dyp_dble = ymin_container_double - dble(yp(ifree_index_ptr))   
!            yp(ifree_index_ptr) = real(ymax_container_double - dyp_dble)
!          end if
!         end if
!         !-- Periodic OB correction Z-Direction --
!         if(iCartTop_periodicity(3))then
!          if(rank_idz.eq.0.and.zp(ifree_index_ptr).gt.zmax_container)then
!            dzp_dble = dble(zp(ifree_index_ptr)) - zmax_container_double    
!            zp(ifree_index_ptr) = real(dzp_dble +zmin_container_double)
!          else if(rank_idz.eq.n_procs_z-1.and.
!      &                       zp(ifree_index_ptr).lt.zmin_container)then
!            dzp_dble = zmin_container_double - dble(zp(ifree_index_ptr))   
!            zp(ifree_index_ptr) = real(zmax_container_double - dzp_dble)
!          end if
!         end if

c        !-- Debug Code------
c        if(rank_id.eq.i_rank_check.and.itime.ge.itime_check)then
c          print*
c          print*,'i, ifree_index_ptr',
c     &            i, ifree_index_ptr
c          print*,'iBPs_top_free_index, iFPs_top_free_index',
c     &            iBPs_top_free_index, iFPs_top_free_index
c          print*,'xp, zp(ifree_index_ptr) ',
c     &    xp(ifree_index_ptr), zp(ifree_index_ptr) 
c          print*,'i_particleType(ifree_index_ptr) ',
c     &            i_particleType(ifree_index_ptr)
c          if(i.le.nb_entering_West)then
c            print*,'iabsolute_BP_Pointer_Index(ifree_index_ptr) ',
c     &              iabsolute_BP_Pointer_Index(ifree_index_ptr) 
c            print*,'nbf_count, nbfm_count, nbfreem_count ',
c     &              nbf_count, nbfm_count, nbfreem_count       
c            print*,'nbf_index, nbfm_index, nbfreem_index ',
c     &              nbf_index, nbfm_index, nbfreem_index       
c          end if
c        end if
c        !-------------------


      end do

c     --- UNPACK DATA ENTERING EAST ---
      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'UnPacking Particle Data entering East'
c        print*,'rank_id ',rank_id
c        print*,'nb_entering_East ',nb_entering_East
c        print*,'np_entering_East ',np_entering_East
c        print*,'nb_local_temp    ',nb_local_temp
c        print*,'nbf_local        ',nbf_local
c        print*,'nbfm_local  ',nbfm_local
c        print*,'nbfreem_local      ',nbfreem_local
c        print*,'nbtemp          ',nbtemp
c        print*,'nptemp          ',nptemp
c        print*,'nbf_temp        ',nbf_temp
c        print*,'nbfm_temp  ',nbfm_temp
c        print*,'nbfreem_temp      ',nbfreem_temp
c      end if
c      !------------------

      nbf_temp       = nbf_temp     + nbf_count
      nbfm_temp      = nbfm_temp    + nbfm_count
      nbfreem_temp   = nbfreem_temp + nbfreem_count
      nFP_temp       = nFP_temp     + nFP_count
      nbf_count      = 0
      nbfm_count     = 0
      nbfreem_count  = 0
      nFP_count      = 0
      
      do i = 1,np_entering_East

        nptemp = nptemp +1

        if(i.le.nb_entering_East)then
          if(iBPs_top_free_index .gt. 0)then
            ifree_index_ptr = iBPs_free_index(iBPs_top_free_index)
            iBPs_top_free_index = iBPs_top_free_index -1
          else if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else!There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nbtemp = nbtemp +1
          ip_index_BP_local(nbtemp) = ifree_index_ptr
          !- Indexing -
          if(nbf_count.lt.no_of_bfPs_entering_East)then   !if(nbf_temp.lt.nbf_local)then
            nbf_count = nbf_count + 1   !nbf_temp = nbf_temp + 1
            nbf_index = nbf_count + nbf_temp
            i_particleType(ifree_index_ptr) = 1   !Fixed Boundary Particle
            ip_index_BP_local_1(nbf_index) = ifree_index_ptr
          else if(nbfm_count.lt.no_of_forcedmBPs_entering_East)then
            nbfm_count = nbfm_count + 1
            nbfm_index = nbfm_count + nbfm_temp
            i_particleType(ifree_index_ptr) = 2   !Forced Moving Boundary Particle
            ip_index_BP_local_2(nbfm_index) = ifree_index_ptr
          else if(nbfreem_count.lt.no_of_freemBPs_entering_East)then
            nbfreem_count = nbfreem_count + 1
            nbfreem_index = nbfreem_count + nbfreem_temp
            i_particleType(ifree_index_ptr) = 3   !Free Moving Boundary Particle
            ip_index_BP_local_3(nbfreem_index) = ifree_index_ptr
          end if
        else
          if(iFPs_top_free_index .gt. 0)then
            ifree_index_ptr = iFPs_free_index(iFPs_top_free_index)
            iFPs_top_free_index = iFPs_top_free_index -1
          else  !There are no free spaces in array other than right at the end
            !ifree_index_ptr = nptemp
            ifree_index_ptr = ihighest_np + 1
            ihighest_np     = ihighest_np + 1
          end if
          nFP_count = nFP_count + 1
          nFP_index = nFP_count + nFP_temp
          ip_index_FP_local(nFP_index) = ifree_index_ptr
          !ip_index_FP_local(nptemp-nb_local_temp) = ifree_index_ptr
          i_particleType(ifree_index_ptr) = 4   !Fluid Particle
        end if
        
        if(iarray_loc_in_use(ifree_index_ptr).ne.0)then
          print*
          print*,'ERROR in Index_check.f: particles entering East'
          print*,'iarray_loc_in_use(ifree_index_ptr).ne.0'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'nb_local ',nb_local
          print*,'nbtemp  ',nbtemp
          print*,'nptemp  ',nptemp
          print*,'ifree_index_ptr ',ifree_index_ptr
          print*,'iarray_loc_in_use(ifree_index_ptr)',
     &            iarray_loc_in_use(ifree_index_ptr)
          stop     
        end if

        xp(ifree_index_ptr)              = P_xyz_data_entering_East(i,1)
c       yp(ifree_index_ptr)              =
        zp(ifree_index_ptr)              = P_xyz_data_entering_East(i,2)
        up(ifree_index_ptr)              = P_data_entering_East(i,1)
c       vp(ifree_index_ptr)              =
        wp(ifree_index_ptr)              = P_data_entering_East(i,2)
        rhop(ifree_index_ptr)            = P_data_entering_East(i,3)
        p(ifree_index_ptr)               = P_data_entering_East(i,4)
        pm(ifree_index_ptr)              = P_data_entering_East(i,5)
        pVol(ifree_index_ptr)            = P_data_entering_East(i,6)
        cs(ifree_index_ptr)              = P_data_entering_East(i,7)
        sum_wab(ifree_index_ptr)         = P_data_entering_East(i,8)
        tau_xx_uno_rho2(ifree_index_ptr) = P_data_entering_East(i,9) 
c       tau_xy_uno_rho2(ifree_index_ptr) =
        tau_xz_uno_rho2(ifree_index_ptr) = P_data_entering_East(i,10)
c       tau_yy_uno_rho2(ifree_index_ptr) =
c       tau_yz_uno_rho2(ifree_index_ptr) =
        tau_zz_uno_rho2(ifree_index_ptr) = P_data_entering_East(i,11)
        TEp(ifree_index_ptr)             = P_data_entering_East(i,12)
        dudx_CSPH(ifree_index_ptr)       = P_data_entering_East(i,13)
c       dudy_CSPH(ifree_index_ptr)       = 
        dudz_CSPH(ifree_index_ptr)       = P_data_entering_East(i,14)
c       dvdx_CSPH(ifree_index_ptr)       = 
c       dvdy_CSPH(ifree_index_ptr)       = 
c       dvdz_CSPH(ifree_index_ptr)       = 
        dwdx_CSPH(ifree_index_ptr)       = P_data_entering_East(i,15)
c       dwdy_CSPH(ifree_index_ptr)       = 
        dwdz_CSPH(ifree_index_ptr)       = P_data_entering_East(i,16)
        drhodx_CSPH(ifree_index_ptr)     = P_data_entering_East(i,17)
c       drhody_CSPH(ifree_index_ptr)     = 
        drhodz_CSPH(ifree_index_ptr)     = P_data_entering_East(i,18)
        dTEdx_CSPH(ifree_index_ptr)      = P_data_entering_East(i,19)
c       dTEdy_CSPH(ifree_index_ptr)      = 
        dTEdz_CSPH(ifree_index_ptr)      = P_data_entering_East(i,20)
        iarray_loc_in_use(ifree_index_ptr) = 1
        iflag(ifree_index_ptr) = 1

        if(i_filterTransfer_Perform.eq.1)then
          rhop_sum(ifree_index_ptr) = rhop_sum_data_entering_East(i)
        endif
        if(i.le.nb_entering_East)then
          iabsolute_BP_Pointer_Index(ifree_index_ptr) = 
     &                           i_ABP_entering_East(i)
        end if
        !- To be used with Verlet algorithm -
        if(i_algorithm.eq.2) then 
          uo(ifree_index_ptr)   = P_verlet_data_entering_East(i,1)
c         vo(ifree_index_ptr)   =
          wo(ifree_index_ptr)   = P_verlet_data_entering_East(i,2)
          rhoo(ifree_index_ptr) = P_verlet_data_entering_East(i,3)
          TEo(ifree_index_ptr)  = P_verlet_data_entering_East(i,4)
        endif

        !-- MPI Periodic OB correction --
        !-- Periodic OB correction X-Direction --
        if(iCartTop_periodicity(1))then
         if(rank_idx.eq.0.and.xp(ifree_index_ptr).gt.xmax_container)then
           dxp_dble = dble(xp(ifree_index_ptr)) - xmax_container_double    
           xp(ifree_index_ptr) = real(dxp_dble +xmin_container_double)
         else if(rank_idx.eq.n_procs_x-1.and.
     &                       xp(ifree_index_ptr).lt.xmin_container)then
           dxp_dble = xmin_container_double - dble(xp(ifree_index_ptr))   
           xp(ifree_index_ptr) = real(xmax_container_double - dxp_dble)
         end if
        end if
!         !-- Periodic OB correction Y-Direction --
!         if(iCartTop_periodicity(2))then
!          if(rank_idy.eq.0.and.yp(ifree_index_ptr).gt.ymax_container)then
!            dyp_dble = dble(yp(ifree_index_ptr)) - ymax_container_double    
!            yp(ifree_index_ptr) = real(dyp_dble +ymin_container_double)
!          else if(rank_idy.eq.n_procs_y-1.and.
!      &                       yp(ifree_index_ptr).lt.ymin_container)then
!            dyp_dble = ymin_container_double - dble(yp(ifree_index_ptr))   
!            yp(ifree_index_ptr) = real(ymax_container_double - dyp_dble)
!          end if
!         end if
!         !-- Periodic OB correction Z-Direction --
!         if(iCartTop_periodicity(3))then
!          if(rank_idz.eq.0.and.zp(ifree_index_ptr).gt.zmax_container)then
!            dzp_dble = dble(zp(ifree_index_ptr)) - zmax_container_double    
!            zp(ifree_index_ptr) = real(dzp_dble +zmin_container_double)
!          else if(rank_idz.eq.n_procs_z-1.and.
!      &                       zp(ifree_index_ptr).lt.zmin_container)then
!            dzp_dble = zmin_container_double - dble(zp(ifree_index_ptr))   
!            zp(ifree_index_ptr) = real(zmax_container_double - dzp_dble)
!          end if
!         end if

c        !-- Debug Code------
c        if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        !if(itime.ge.itime_check.and.rank_id.eq.3)then
c          print*,'i, ifree_index_ptr',
c     &            i, ifree_index_ptr
c          print*,'iBPs_top_free_index, iFPs_top_free_index',
c     &            iBPs_top_free_index, iFPs_top_free_index
c          print*,'xp, zp(ifree_index_ptr) ',
c     &            xp(ifree_index_ptr), zp(ifree_index_ptr) 
cc          print*,'xp(ifree_index_ptr)   ',
cc     &            xp(ifree_index_ptr)
cc          print*,'zp(ifree_index_ptr)   ',
cc     &            zp(ifree_index_ptr)
cc          print*,'up(ifree_index_ptr)   ',
cc     &            up(ifree_index_ptr)
cc          print*,'wp(ifree_index_ptr)   ',
cc     &            wp(ifree_index_ptr)
cc          print*,'rhop(ifree_index_ptr) ',
cc     &            rhop(ifree_index_ptr)
cc          print*,'p(ifree_index_ptr)    ',
cc     &            p(ifree_index_ptr)
cc          print*,'pm(ifree_index_ptr)   ',
cc     &            pm(ifree_index_ptr)
cc          print*,'pVol(ifree_index_ptr) ',
cc     &            pVol(ifree_index_ptr)
cc          print*,'cs(ifree_index_ptr)   ',
cc     &            cs(ifree_index_ptr)
c          print*,'i_particleType(ifree_index_ptr) ',
c     &            i_particleType(ifree_index_ptr)
c         if(i.le.nb_entering_East)then
c            print*,'iabsolute_BP_Pointer_Index(ifree_index_ptr) ',
c     &              iabsolute_BP_Pointer_Index(ifree_index_ptr) 
c            print*,'nbf_count, nbfm_count, nbfreem_count ',
c     &              nbf_count, nbfm_count, nbfreem_count       
c            print*,'nbf_index, nbfm_index, nbfreem_index ',
c     &              nbf_index, nbfm_index, nbfreem_index       
c          end if
c        end if
c        !-----------------------

      end do
      
      if(nbtemp.ne.nb_local_temp)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'nbtemp.ne.nb_local_temp'
        stop        
      end if
      if(nptemp.ne.np_local_temp)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'nptemp.ne.np_local_temp'
        stop        
      end if

      nb_local = nbtemp
      np_local = nptemp
      nbfp1_local = nbf_local + 1
      nbp1_local  = nb_local  + 1
      if(iBC.eq.1)then
         nstart=nbp1_local
      else
         nstart=1
      endif
      nstart_minus1 = nstart - 1
      
      if(nptemp > ihighest_np) ihighest_np = np_local

      !call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point

c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'All Data Entering UnPacked'
c        print*,'itime ',itime
c        print*,'rank_id ',rank_id
c        print*,'nb_entering_East ',nb_entering_East
c        print*,'np_entering_East ',np_entering_East
c        print*,'nb_entering_West ',nb_entering_West
c        print*,'np_entering_West ',np_entering_West
c        print*,'nbf_local          ',nbf_local
c        print*,'nbfm_local    ',nbfm_local
c        print*,'nbfreem_local        ',nbfreem_local
c        print*,'nb_local           ',nb_local
c        print*,'np_local           ',np_local
c        print*
c        print*,'Performing Non-MPI Periodic Boundary Correction'
c        print*
c      end if
c      !--------------------

      !-- Periodic Open Boundary Correction --
      do kpcount=1,np_local
        if(kpcount.le.nb_local)then
          k = ip_index_BP_local(kpcount)
        else
          k = ip_index_FP_local(kpcount-nb_local)
        end if
        !-- Periodic OB correction X-Direction --
        if(i_periodicOBs(1).eq.1)then
          if(xp(k).gt.xmax_container)then
            dxp_dble = dble(xp(k)) - xmax_container_double 
            xp(k) = real(dxp_dble +xmin_container_double)
          else if(xp(k).lt.xmin_container)then
            dxp_dble = xmin_container_double - dble(xp(k))   
            xp(k) = real(xmax_container_double - dxp_dble)
          end if
        end if
c        !-- Periodic OB correction Y-Direction --
c        if(i_periodicOBs(2).eq.1)then
c          if(yp(k).gt.ymax_container)then
c            dyp_dble = dble(yp(k)) - ymax_container_double 
c            yp(k) = real(dyp_dble +ymin_container_double)
c          else if(yp(k).lt.ymin_container)then
c            dyp_dble = ymin_container_double - dble(yp(k))   
c            yp(k) = real(ymax_container_double - dyp_dble)
c          end if
c        end if
        !-- Periodic OB correction Z-Direction --
        if(i_periodicOBs(3).eq.1)then
          if(zp(k).gt.zmax_container)then
            dzp_dble = dble(zp(k)) - zmax_container_double 
            zp(k) = real(dzp_dble +zmin_container_double)
          else if(zp(k).lt.zmin_container)then
            dzp_dble = zmin_container_double - dble(zp(k))   
            zp(k) = real(zmax_container_double - dzp_dble)
          end if
        end if
      end do

c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'Non-MPI Periodic OB correction Complete'
c        print*,'itime ',itime
c        print*,'rank_id ',rank_id
c        print*
c        print*,'Checking Particle Index Pointers'
c        print*
c        print*
c      end if
c      !-------------------

      !ihighest_np = np_local  !only use this for periodic boundaries
c     ..

c     ***  reorder the ip_index  for the next loop  ***
      !This ensures that after particles have left,
      !the linked list when accessed sequentially
      !only looks at array locations that contain particles
      !-- Serial Version --
      !ip_index_no = nb
      !do k = nbp1,ihighest_np
      !   if(iarray_loc_in_use(k).eq.1)then
      !      ip_index_no = ip_index_no+1
      !      ip_index(ip_index_no) = k
      !   end if
      !end do
      !-- MPI Version --
      ip_index_no_BPs = 0
      ip_index_no_FPs = 0
      kk_highest = 0
      ip_index_no_BPs_1 = 0
      ip_index_no_BPs_2 = 0
      ip_index_no_BPs_3 = 0
      do k = 1,ihighest_np
        if(iarray_loc_in_use(k).eq.1)then
         if((xp(k).lt.xmin_rank.or.xp(k).gt.xmax_rank))then
          print*
          print*,'Problem with particle location in'
     &                //' index_check_EastWest.f'
          print*,'itime ',itime
          print*,'rank_id ',rank_id
          print*,'rank_idx, rank_idz ',
     &            rank_idx, rank_idz
          print*,'nb_local ',nb_local
          print*,'np_local ',np_local
          print*,'xmin_rank ',xmin_rank
          print*,'xmax_rank ',xmax_rank
          print*,'zmin_local ',zmin_local
          print*,'k ',k
          print*,'xp(k), zp(k) ',xp(k), zp(k)
          print*,'i_particleType(k)    ',i_particleType(k)
          print*,'iarray_loc_in_use(k) ',iarray_loc_in_use(k)
          stop
         end if
          if(i_particleType(k).lt.4)then  !Boundary Particle
            ip_index_no_BPs = ip_index_no_BPs+1
            ip_index_BP_local(ip_index_no_BPs) = k
            if(i_particleType(k).eq.1)then
              ip_index_no_BPs_1 = ip_index_no_BPs_1 + 1
              ip_index_BP_local_1(ip_index_no_BPs_1) = k
            else if(i_particleType(k).eq.2)then
              ip_index_no_BPs_2 = ip_index_no_BPs_2 + 1
              ip_index_BP_local_2(ip_index_no_BPs_2) = k
            else if(i_particleType(k).eq.3)then
              ip_index_no_BPs_3 = ip_index_no_BPs_3 + 1
              ip_index_BP_local_3(ip_index_no_BPs_3) = k
            else  !Error catching
              print*,'Problem #1 with i_particleType(k) in '
     &               //' index_check_EastWest.f'
              print*,'itime ',itime
              print*,'rank_id ',rank_id
              print*,'k ',k
              print*,'xp(k), zp(k) ',xp(k), zp(k)
              print*,'i_particleType(k) ',i_particleType(k)
              stop
            end if
          else if(i_particleType(k).eq.4)then  !Fluid Particle
            ip_index_no_FPs = ip_index_no_FPs+1
            ip_index_FP_local(ip_index_no_FPs) = k
          else  !Error catching
            print*,'Problem #2 with i_particleType(k) in '
     &             //' index_check_EastWest.f'
            print*,'itime ',itime
            print*,'rank_id ',rank_id
            print*,'k ',k
            print*,'xp(k), zp(k) ',xp(k), zp(k)
            print*,'i_particleType(k) ',i_particleType(k)
            stop
          end if
          if(k.gt.kk_highest) kk_highest = k   !Used for starting the imported image particles count
        end if
      end do
      if(kk_highest.gt.ihighest_np)then
        print*
        print*,'ERROR in Index_check.f'
        print*,'itime ',itime
        print*,'rank_id ',rank_id
        print*,'kk_highest.gt.ihighest_np'
        print*,'kk_highest  ',kk_highest
        print*,'ihighest_np ',ihighest_np
        print*
        stop        
      endif
      kk_highest = ihighest_np + 1
      
c      !- Image Printout to Screen- 
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c      !if(itime.ge.itime_check.and.rank_id.eq.2)then
c        print*
c        print*,'rank_id ',rank_id
c        print*,'itime   ',itime
c        print*,'nbf_local       ' ,nbf_local 
c        print*,'nbfm_local ' ,nbfm_local 
c        print*,'nbfreem_local     ' ,nbfreem_local 
c        print*,'nb_local       ',nb_local
c        print*,'np_local       ',np_local
c        do ipcount=1,np_local
c          if(ipcount.le.nbf_local)then
c            i = ip_index_BP_local_1(ipcount)
c          else if(ipcount.le.nbfm_local)then
c            i = ip_index_BP_local_2(ipcount-nbf_local)
c          else if(ipcount.le.nbfreem_local)then
c            i = ip_index_BP_local_3(ipcount-nbfm_local)
c          else
c            i = ip_index_FP_local(ipcount-nb_local)
c          end if
c          print*,'ipcount -> i, Type, xp, zp',
c     &                ipcount,i,i_particleType(i),xp(i),zp(i)
c        end do
c      endif
     
      if(i_MPI_Adapt.eq.1.and.n_procs_x.gt.1.and.n_procs_z.eq.1)then
        if(itime/n_itime_Adapt*n_itime_Adapt.eq.itime
     &                      .or.itime.eq.itime_init)then
          call Adapt_MPI_Regrid
        endif
      endif
      
      !-  Update indices & FLOATING BODY distances from Body Centre -
      if(num_FB.gt.0)then
        !- Clear indexing system
        !do i_num_FB = 1,num_FB
        !  nb_FB_local(i_num_FB) = nbfm_local
        !enddo
        do ipcount = nbfm_local+1,nbfreem_local
          i = ip_index_BP_local_3(ipcount - nbfm_local)
          i_absolute = iabsolute_BP_Pointer_Index(i)
          i_num_FB = i_FB_Pointer_Info(i_absolute)
          !- Error catching -
          if(i_num_FB.lt.1)then
            print*,'ERROR in index_check_2D.f'
            print*,'i_num_FB.lt.1'
            print*,'rank_id, itime ',rank_id, itime
            print*,'nbfm_local+1 ',nbfm_local+1
            print*,'nbfreem_local       ',nbfreem_local
            print*,'ipcount, i     ',ipcount, i
            print*,'i_absolute     ',i_absolute
            print*,'i_num_FB       ',i_num_FB
            stop
          endif
c          xp_temp = BP_xyz_Data(i_absolute,1)
cc         yp_temp =          
c          zp_temp = BP_xyz_Data(i_absolute,2)
c          xp_minus_R0(i) = xp_temp - Box_XC(i_num_FB)
cc         yp_minus_R0(i) = 
c          zp_minus_R0(i) = zp_temp - Box_ZC(i_num_FB)
        enddo
      endif


c      !- Image Printout to Screen- 
c      if(itime.ge.itime_check.and.rank_id.eq.3)then
c      !if(itime.ge.itime_check.and.rank_id.eq.2)then
c        print*
c        print*,'rank_id ',rank_id
c        print*,'itime   ',itime
c        print*,'nb_local       ',nb_local
c        print*,'np_local       ',np_local
c        do ipcount=1,np_local
c          !if(ipcount.le.nb_local)then
c          !  i = ip_index_BP_local(ipcount)
c          !else
c          !  i = ip_index_FP_local(ipcount-nb_local)
c          !end if
c          if(ipcount.le.nbf_local)then
c            i = ip_index_BP_local_1(ipcount)
c          else if(ipcount.le.nbfm_local)then
c            i = ip_index_BP_local_2(ipcount-nbf_local)
c          else if(ipcount.le.nbfreem_local)then
c            i = ip_index_BP_local_3(ipcount-nbfm_local)
c          else
c            i = ip_index_FP_local(ipcount-nb_local)
c          end if
c          print*,'ipcount -> i, Type, xp, zp',
c     &                ipcount,i,i_particleType(i),xp(i),zp(i)
c        end do
c        print*,'kk_highest, ihighest_np',kk_highest, ihighest_np
c        print*,'ip_index_no_BPs ,ip_index_no_FPs ',
c     &          ip_index_no_BPs ,ip_index_no_FPs
c        no_array_positions = 0
c        !do ipcount = 1,nb_local
c        !  i = ip_index_BP_local(ipcount)
c        !enddo
c        do kk = 1,ihighest_np
c          print*,'kk, iarray_loc_in_use(kk), xp, zp ', 
c     &            kk, iarray_loc_in_use(kk),xp(kk),zp(kk)
c          if(iarray_loc_in_use(kk).eq.1)then
c            no_array_positions = no_array_positions +1
c          endif
c        enddo
c        print*,'no_array_positions ',no_array_positions
c        print*,'nb_local        ',nb_local
c        print*,'np_local        ',np_local
c        print*,'iBPs_top_free_index ',iBPs_top_free_index
c        do i=1,iBPs_top_free_index
c          print*,'i,iBPs_free_index(i) ',i,iBPs_free_index(i)
c        enddo
c        print*,'iFPs_top_free_index ',iFPs_top_free_index
c        do i=1,iFPs_top_free_index
c          print*,'i,iFPs_free_index(i) ',i,iFPs_free_index(i)
c        enddo        
c        if(net_np_change.gt.0)then
c          print*,'ifree_index_ptr     ',ifree_index_ptr
c        endif
c        !stop
c      end if

c      !-- Debug Code------
c      if(itime.ge.itime_check)then
c        do i=0,n_procs-1
c          if(rank_id.eq.i)then   !if(i.eq.-rank_id-1)then   !if(i.eq.rank_id)then
c            print*
c            print*,'itime ',itime
c            write(*,*),'rank_id ', rank_id
c            print*,'nbf_local ',nbf_local
c            print*,'nb_local  ',nb_local
c            print*,'nb_local   ',nb_local
c            print*,'np_local   ',np_local
c          end if
c          call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
c        end do
c      end if
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then   !if(rank_id.eq.-10)then
c        print*
c        print*,'index_check finished'
c        print*,'itime ',itime
c        print*,'rank_id ',rank_id
c        !stop
c      end if
c      !---------------------

c     .. Inflow Periodic OB correction .. 
      if(iCartTop_periodicity(1).and.rank_id.eq.-10)then
        if(rank_id.eq.-10)then
          print*
          print*,'Performing Inflow Periodic OB correction'
        end if      
        n_inlet = 0
        flow_rate = 0.0
        flow_in_old_temp = flow_in_old
        do ipcount=nb_local+1,np_local
          i = ip_index_FP_local(ipcount-nb_local)
          if(xp(i).gt.xmin_rank.and.xp(i).lt.xmin_rank + 2.0*h)then
            n_inlet = n_inlet + 1
            flow_rate = flow_rate + up(i)           
c            print*,'n_inlet, xp(i), up(i), flow_rate',
c     &              n_inlet, xp(i), up(i), flow_rate
          end if
        end do
        flow_rate = flow_rate / n_inlet
        flow_force = (2.0*(flow_rate - U_in) - (flow_in_old - U_in))
     &             /(2.0*dt)
        grx = - flow_force
        if(itime.lt.1000)then
          grx = grx*real(itime)/real(1000)
        end if
        flow_in_old = flow_rate
c        print*,'rank_id,itime,U_in,flow_in_old,flow_rate,flow_force,
c     &n_inlet'
c        write(*,*)rank_id,itime,U_in,flow_in_old,flow_rate,
c     &            flow_force,n_inlet

        call MPI_BCast(grx, 1, MPI_REAL, 
     &               0, MPI_COMM_WORLD, ierr)
        if(rank_id.eq.-10)then
          !print*
          write(*,366),itime, rank_id, n_inlet,
     &      grx, U_in,flow_in_old_temp,flow_rate,flow_force
          print*
        end if
      end if
366   format(i6,i3,i5,6(' ',e12.6))


c      if(itime.ge.itime_check.and.rank_id.ge.0)then
c        print*
c        print*,'END OF index_check.f'
c        print*,'itime, rank_id ',itime, rank_id
c        print*,'np_local ',np_local
c        if(itime.eq.0)stop
c      end if

      !stop
c
800   return
      end !subroutine
