c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.


      subroutine monaghanBC(iip,jjp,drx,drz,dot,
     &                      duxp,duzp,
     &                      fxbp,fzbp)
c
      include 'common.2D'

c     Boundary condition JWPCOE, 32
c
c     iip is a WaterParticle (WP)
c     jjp is a BoundaryParticle (BP)
c
c     drx = xp(WP) - xp(BedP)
c     drz = zp(WP) - zp(BedP)


c      Projection of WP position onto normal
       pdot3 = drx*xnb(jjp)+drz*znb(jjp)

c       if(pdot3.gt.0.0)then  !ensures that WP is on correct side of boundary
c
c         xpp = magnitude(vec(dr) x vec(n))   cross-product
c
c         xpp=(drx*ynb(jjp)-dryp*xnb(jjp))


c        Projection of WP on plane of boundary
c
c        r(BP->WP) = [drx,dryp,drz]
c
c        r(BP->WP projection) = r(BP->WP) -[r(BP->WP).n/(|n|.|n|)]n
c
c        normal to plane = n
c        magnitude of normal = |n| = 1

         drx_proj = drx - (pdot3)*xnb(jjp)
         drz_proj = drz - (pdot3)*znb(jjp)


c        Projection of plane-projected vector r(BP->WP projection) onto tangent

         xpp = drx_proj*xtb(jjp)+drz_proj*ztb(jjp)


         if(xpp.gt.0.0)then
           deltappt = deltaptb(jjp,2)
         else
           deltappt = deltaptb(jjp,1)
         end if
c      
c
c         
c         
c         
               
         !- Correction for particle exactly deltappt from another particle -
         g_dot_t = grx*xtb(jjp) + grz*ztb(jjp)
         xpp_orig = xpp
         if(.NOT.(abs(xpp).lt.deltappt))then  !i.e. >=deltappt
           if(g_dot_t.gt.0.0)then
             if(xpp.gt.0.0)then
               xpp = 0.99*xpp
             endif
           else   !if(g_dot_t.lt.0.0)then
             if(xpp.lt.0.0)then
               xpp = 0.99*xpp
             endif
           endif
         elseif(.NOT.(abs(xpp).gt.0.0))then  !i.e. ==0.0
           if(g_dot_t.gt.0.0)then
             xpp = xpp - 0.01*deltappt
           else   !if(g_dot_t.lt.0.0)then
             xpp = xpp + 0.01*deltappt
           endif
         endif

         Prx = abs(xpp)/deltappt    
c        Pry =          
         facp = 0.0
         f_tb = 0.0
         f_fb = 0.0
         chi_X = 0.0
c        chi_Y         
         Q_fn = 0.0
         i_CornerParticle = 0
         
         !normalized distance from wall
         if(pdot3.gt.0.0)then ! particle inside
             q=pdot3 * one_over_2h           !normalized distance from wall
         else if(pdot3.lt.0.0)then !Particle ouside 
             q=0.2*abs(pdot3) * one_over_2h  !normalized distance from wall
         else !Particle on the boundary
             q=h*(1.0e-3) * one_over_2h      !correcting for zero distance from wall
         end if

         if(q.lt.1.0.and.pdot3.gt.0.0)then
     
     
c            u_normal < 0 if particles are approaching
c            u_normal > 0 if particles are moving apart
      
           !-- Corner Particle Identification --
           jjp_absolute = iabsolute_BP_Pointer_Index(jjp)
           j_minus1 = iBP_Pointer_Info(jjp_absolute,3)     !Absolute index of i-1 neighbour BP jjp
           j_plus1  = iBP_Pointer_Info(jjp_absolute,4)     !Absolute index of i+1 neighbour BP jjp
           
           !- Calculation of normal velocity, Repulsive singularity function (Q_fn) -
           !- and tangential distance reduction function chi_X                      -
           if(xpp.lt.0.0.and.jjp_absolute.eq.j_minus1)then
             !Identify partner-CornerParticle index
             xp_BP_jminus1 = BP_xyz_Data(j_minus1,1)
c            yp_BP_jminus1 =
             zp_BP_jminus1 = BP_xyz_Data(j_minus1,2)             
             mod_xp_j = abs(xp_BP_jminus1-xp(jjp))
c            mod_yp_j = abs(yp_BP_jminus1-yp(jjp))
             mod_zp_j = abs(zp_BP_jminus1-zp(jjp))
             if((mod_xp_j.lt.h_over_100).and.
     &          (mod_zp_j.lt.h_over_100))then
               jjp_minus1 = jjp_absolute - 1
             else
               jjp_minus1 = j_minus1
             endif
             if(jjp_minus1.ne.j_minus1)then
               drp = sqrt(drx*drx + drz*drz)
               one_over_drp = 1.0/(drp + h_over_1000)
               q_r = drp*one_over_2h
               if(q_r.lt.1.0)then
                 Q_fn = (1.0-q_r)/sqrt(q_r)
               else
                 Q_fn = 0.0
               end if
               u_normal = (duxp*drx+duzp*drz)*one_over_drp
               chi_X = pdot3*one_over_drp
c              chi_Y               
c               -- Serial version--
c               vn1_dot_n2 =xnb(jjp)*xnb(jjp_minus1)
c     &                   + znb(jjp)*znb(jjp_minus1)
               vn1_dot_n2 =xnb(jjp)*xnb(jjp -1)   !jjp - 1 I KNOW IS NOT CORRECT for MPI!
     &                   + znb(jjp)*znb(jjp -1)   !I need to know the indices of the other CP!!!
               vn1_dot_rij = (xnb(jjp)*drx + znb(jjp)*drz)*one_over_drp
               if(vn1_dot_rij.lt.vn1_dot_n2)then  !Activate only for WPs within arc of Corner Particles normals
                 Q_fn = 0.0
               end if
               xtb_r =  drz*one_over_drp
c              ytb_r =  
               ztb_r = -drx*one_over_drp               
c
c
c               
             else
               Q_fn = (1.0-q)/sqrt(q)
               u_normal = duxp*xnb(jjp) + duzp*znb(jjp)
               chi_X = (1.0 - Prx)
c              chi_Y               
               q_r = q
               xtb_r = xtb(jjp)
c              ytb_r =  
               ztb_r = ztb(jjp)
c
c
c               
             endif
             i_CornerParticle = 1   !Activated as a Corner Particle
           else if(xpp.ge.0.0.and.jjp_absolute.eq.j_plus1)then
             !Identify partner-CornerParticle index
             xp_BP_jplus1 = BP_xyz_Data(j_plus1,1)
c            yp_BP_jplus1 =
             zp_BP_jplus1 = BP_xyz_Data(j_plus1,2)
             mod_xp_j = abs(xp_BP_jplus1-xp(jjp))
c            mod_yp_j = abs(yp_BP_jplus1-yp(jjp))
             mod_zp_j = abs(zp_BP_jplus1-zp(jjp))
             if((mod_xp_j.lt.h_over_100).and.
     &          (mod_zp_j.lt.h_over_100))then
               jjp_plus1 = jjp_absolute + 1
             !elseif(jjp.gt.nbfm)then  <--- for floating bodies only
             else
               jjp_plus1 = j_plus1
             endif
             if(jjp_plus1.ne.j_plus1)then
               drp = sqrt(drx*drx + drz*drz)
               one_over_drp = 1.0/(drp + h_over_1000)
               q_r = drp*one_over_2h
               if(q_r.lt.1.0)then
                 Q_fn = (1.0-q_r)/sqrt(q_r)
               else
                 Q_fn = 0.0
               end if
               u_normal = (duxp*drx+duzp*drz)*one_over_drp
               chi_X = pdot3*one_over_drp
c              chi_Y               
c               -- Serial version--
c               vn1_dot_n2 = xnb(jjp)*xnb(jjp_plus1)
c     &                    + znb(jjp)*znb(jjp_plus1)
               vn1_dot_n2 = xnb(jjp)*xnb(jjp+1)   !jjp + 1 I KNOW IS NOT CORRECT for MPI!
     &                    + znb(jjp)*znb(jjp+1)   !I need to know the indices of the other CP!!!
               vn1_dot_rij = (xnb(jjp)*drx + znb(jjp)*drz)*one_over_drp
               if(vn1_dot_rij.lt.vn1_dot_n2)then  !Activate only for WPs within arc of Corner Particles normals
                 Q_fn = 0.0
               end if
               xtb_r =  drz*one_over_drp
c              ytb_r =  
               ztb_r = -drx*one_over_drp
             else
               Q_fn = (1.0-q)/sqrt(q)
               u_normal = duxp*xnb(jjp) + duzp*znb(jjp)
               chi_X = (1.0 - Prx)
c              chi_Y               
               q_r = q
               xtb_r = xtb(jjp)
c              ytb_r =  
               ztb_r = ztb(jjp)
             endif
             i_CornerParticle = 1   !Activated as a Corner Particle
           else if(abs(Prx).lt.1.0)then  !Acts as a Conventional Boundary Particle
             Q_fn = (1.0-q)/sqrt(q)
             u_normal = duxp*xnb(jjp) + duzp*znb(jjp)
             chi_X = (1.0 - Prx)
c            chi_Y               
             q_r = q
             xtb_r = xtb(jjp)
c            ytb_r =  
             ztb_r = ztb(jjp)
           end if
             
           if(chi_X.gt.0.0.and.Q_fn.gt.0.0)then
             !c0 = coeff.Vmax (Vmax = sqrt(g.h_SWL); coef = beta_coef.10, beta_coef2 = beta_coef**2
             !beta_coef2 = Factor that B is changed by 
             if(u_normal.gt.0.0)then
              App=(0.01*cs(iip)*cs(iip)*one_over_beta_coef2
     &                -cs(iip)*u_normal*one_over_beta_coef)*one_over_h
             else
              App=0.01*(cs(iip)*cs(iip)*one_over_beta_coef2)*one_over_h    
             endif

       
c          --- Depth Correction ---
             eps_min = 0.02
             epsilon_Z = abs(1.0 - (zp(jjp)/h_SWL))
             dyn_head =(up(jjp)*up(jjp) + wp(jjp)*wp(jjp))
     &                                    /(2.0*abs(grz)*h_SWL)
           
             if(epsilon_Z.gt.1.0-eps_min)then !Close to the bottom (No correction )
                epsilon_Z = 1.0
             elseif(zp(iip).gt.h_SWL)then !over still water level          
                epsilon_Z = eps_min + dyn_head
             else                 !under still water level & far from bottom
                epsilon_Z = eps_min + epsilon_Z + dyn_head
             endif

c          --- Velocity Correction ---
             if(u_normal.gt.0.0)then
                epsilon_dyn = 0.0
             else
                epsilon_dyn = 40.*abs(u_normal)/cs0 
c                epsilon_dyn = min(1.0,40.*abs(u_normal)/cs0)     !factor 20 needed   
c                                                                    ! to keep scheme stable                                                                                  ! with Monaghan BC wavemaker
             endif

             if(epsilon_dyn.gt.1.0)then
                epsilon_dyn = 1.0
             endif

             Appc = App*(epsilon_Z + epsilon_dyn) !Force correction
             
             fmax = Appc*Q_fn  !*vnorm_mass
             facp = chi_X*fmax
             
             fxbp = xnb(jjp)*facp
             fzbp = znb(jjp)*facp   

             !- Angular Momentum Correction -
c               g_dot_t = grx*xtb_r + grz*ztb_r  !Be careful of this:  g_dot_t = grx*xtb(jjp) + grz*ztb(jjp)
c
c               g_dot_t_normalised = abs(g_dot_t)/abs(grz)
c               if(xpp.gt.0.0)then
c                 f_tb =  two_alpha_h*g_dot_t_normalised*fmax*
c     &                   (0.25*deltappt*one_over_r_ij_dot_n)
c               else
c                 f_tb = -two_alpha_h*g_dot_t_normalised*fmax*
c     &                   (0.25*deltappt*one_over_r_ij_dot_n)
c               end if
               f_tb =  two_alpha_h_over_4*fmax
     &                   *one_over_Q_of_deltax

               !- Correction for WP near bottom corner or surface
               h_SWL_minus_zp = h_SWL - zp(iip)
               !rNum_h = 3.0*h
               !one_over_rNum_h = 1.0/rNum_h
               f_tb_orig = f_tb
               if(h_SWL_minus_zp.lt.0.0)then  !WP is above h_SWL
                 f_tb = 0.0  !*f_tb
               elseif(h_SWL_minus_zp.lt.rNum_h)then
                 epsilon_Z_AMC = h_SWL_minus_zp * one_over_rNum_h
                 if(epsilon_Z_AMC.lt.0.90)then
                   epsilon_Z_AMC = (epsilon_Z_AMC + 0.10)
                 endif
                 f_tb = f_tb * epsilon_Z_AMC
               !elseif(zp(iip).lt.rNum_h)then  !i.e. less than 2h above the bottom
               !  epsilon_Z_AMC = zp(iip) * one_over_rNum_h
               !  f_tb = f_tb * epsilon_Z_AMC
               endif
               
               if(xpp.gt.0.0)then
                 sign_rij_dot_tj =   1.0
               else
                 sign_rij_dot_tj =  -1.0
               end if
c              if(ypp.gt.0.0)then
c                sign_rij_dot_sj =   1.0
c              else
c                sign_rij_dot_sj =  -1.0
c              end if
               f_tb_t = f_tb !* g_dot_t_normalised
     &                       * sign_rij_dot_tj !* deltappt



               fxbp = fxbp + xtb_r*f_tb_t
               fzbp = fzbp + ztb_r*f_tb_t
               

c               if(rank_id.eq.-10.and.(iip.eq.12875.or.iip.eq.12929))then
c                 print*
c                 print*,'in MonBC, itime = ',itime
c                 print*,'iip, jjp ',iip,jjp
c                 print*,'xp(iip), zp(iip) ',xp(iip), zp(iip)
c                 print*,'xp(jjp), zp(jjp) ',xp(jjp), zp(jjp)
c                 print*,'drx, drz ',drx,drz
c                 print*,'xnb, znb ',xnb(jjp),znb(jjp)
c                 print*,'xpp, deltappt ',xpp, deltappt
c                 if(abs(xpp).lt.deltappt)then
c                  print*,'epsilon_Z',epsilon_Z
c                  print*,'epsilon_dyn',epsilon_dyn
c                  print*,'App ',App
c                  print*,'Appc',Appc
c                  print*,'facp',facp
c                  print*,'App*(1-q)/sqrt(q)',App*(1.0-q)/sqrt(q)
c                  print*,'facp ',facp
c                  print*,'fxbp ',fxbp
c                  print*,'fzbp ',fzbp
c                  !print*
c                  !read(*,*)
c                 endif
c                 print*,'two_alpha_h ',two_alpha_h
c                 print*,'g_dot_t     ',g_dot_t
c                 print*,'one_over_r_ij_dot_n  ',one_over_r_ij_dot_n
c                 print*,'deltappt    ',deltappt
c                 print*,'fmax        ',fmax
c                 print*,'f_tb_orig   ',f_tb_orig
c                 print*,'h_SWL          ',h_SWL
c                 print*,'h_SWL_minus_zp ',h_SWL_minus_zp
c                 print*,'rNum_h          ',rNum_h
c                 print*,'one_over_rNum_h ',one_over_rNum_h
c                 print*,'epsilon_Z_AMC  ',epsilon_Z_AMC
c                 print*,'f_tb        ',f_tb
c                 print*,'xtb(jjp)*f_tb ',xtb(jjp)*f_tb
c                 print*,'ztb(jjp)*f_tb ',ztb(jjp)*f_tb
c                 write(*,858)'ztb(jjp)*f_tb  ',ztb(jjp)*f_tb 
c858              format(a16,' ',e24.16)        
c                 print*
c               endif
             !endif
           else  
             fxbp = 0.0
             fzbp = 0.0
             
           endif   !End of:  if(chi_X.gt.0.0.and.Q_fn.gt.0.0)then

         else
             fxbp = 0.0
             fzbp = 0.0
             
         end if

c      !- Debugging -
c      fxbp = 0.0
c      fzbp = 0.0

c      if(1.eq.2.and.
c     &   itime.le.1.and. 
cc     & (iip.eq.3455))then
cc     & (iip.eq.6302))then
cc     & (iip.eq.6367))then
c     & (iip.eq.2139.and.rank_id.eq.0))then
cc         if(q.lt.1.0.and.q.gt.0.0.and.
cc     &      abs(xpp).lt.deltappt)then
cc         if(q.lt.1.0.and.q.gt.0.0)then
c               print*
c               print*,'itime ',itime
c               print*,'rhop(iip), cs(iip) ',rhop(iip), cs(iip)
c               print*,'iip, xp, zp ',iip,xp(iip),zp(iip)
c               print*,'jjp, xp, zp ',jjp,xp(jjp),zp(jjp)
c               print*,'drx, drz  ',drx,drz
c               print*,'drx_proj, drz_proj ',drx_proj, drz_proj
c               print*,'xpp_orig, xpp, deltappt ',xpp_orig, xpp, deltappt
c               print*,'abs(xpp).lt.deltappt ',abs(xpp).lt.deltappt
c               print*,'q  , q_r, pdot3  ',q,q_r,pdot3
c               print*,'up, wp      ',up(iip),wp(iip)
c               print*,'up, wp      ',up(jjp),wp(jjp)
c               print*,'iip, xnb, znb, xtb, ztb ',
c     &                 xnb(iip),znb(iip),xtb(iip),ztb(iip)
c               print*,'jjp, xnb, znb, xtb, ztb ',
c     &                 xnb(jjp),znb(jjp),xtb(jjp),ztb(jjp)
c               print*,'xtb_r, ztb_r ',xtb_r,ztb_r
c               print*,'abs_t_dot_t ',abs_t_dot_t
c               print*,'g_dot_t            ',g_dot_t
c               print*,'g_dot_t_normalised ',g_dot_t_normalised
c               print*,'u_normal    ',u_normal
c               print*,'u_parallel  ',u_parallel
c               print*,'pmu         ',pmu
c               print*,'fxbp, fzbp  ',fxbp, fzbp
c               print*,'facp, f_tb  ',facp,f_tb
c               print*,'Prx,chi_X,Q_fn,fmax ',Prx,chi_X,Q_fn,fmax
c               print*,'Appc, vnorm_mass    ',Appc, vnorm_mass
c               print*,'App, epsilon_Z, epsilon_dyn ',
c     &                 App, epsilon_Z, epsilon_dyn
c               !print*,'sign(1.0,u_parallel) ',sign(1.0,u_parallel)
c               print*,'sign_u_parallel      ',sign_u_parallel
c               print*,'xnb(jjp)*,znb(jjp)*facp ',
c     &                 xnb(jjp)*facp, znb(jjp)*facp
c               print*,'xtb_r*f_tb, ztb_r*f_tb ',
c     &                 xtb_r*f_tb, ztb_r*f_tb
c               print*,'f_tb_orig, epsilon_Z_AMC  ',
c     &                 f_tb_orig, epsilon_Z_AMC
c               print*,'jjp_absolute, j_minus1, j_plus1 ',
c     &                 jjp_absolute, j_minus1, j_plus1
c               print*,'i_CornerParticle ',i_CornerParticle
c               if(i_CornerParticle.gt.0)then
c                 print*,'drp, one_over_drp ',drp, one_over_drp
c                 print*,'1./drp, h_over_1000 ',1.0/drp, h_over_1000
c                 print*,
c     &               ' jjp_minus1, jjp_plus1 ',
c     &                 jjp_minus1, jjp_plus1
c                 print*,'vn1_dot_rij, vn1_dot_n2 ',
c     &                   vn1_dot_rij, vn1_dot_n2
c               endif
c         print*
c         !read(*,*)
cc        end if
c        if(itime.eq.2)stop
c      end if

      end	
