c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.


      subroutine step
c
      include 'common.2D'

      dt2 = 0.5*dt

c	Predictor- Corrector Method

c  ...  compute acceleration
c

c      !-- Debug Code------
c      itime_check = 0    !10e8   !0    !
c      i_rank_check = 0
c      i_ParticleCheck = 537  !2366  !nb_local + 1
      
      if(rank_id.eq.0.and.itime/100*100 .eq. itime) then
        write(*,*) 'In step:   time = ',time,' itime = ',itime
      endif

       nb_image_local      = nb_local
       np_image_local      = np_local
       nbf_image_local     = nbf_local
       nbfm_image_local    = nbfm_local
       nbfreem_image_local = nbfreem_local
       nbp1_local = nb_local + 1
       nfbp1_local = nbf_local + 1     

c      !- Diagnostic Info -
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'In step_predictor_corrector.f BEFORE anything'
c        i =  i_ParticleCheck
c        print*,'i, xp, zp ',i,xp(i),zp(i)
c        print*,'up, wp    ',up(i),wp(i)
c        print*,'rhop(i) ',rhop(i)
c        print*,'p(i)    ',p(i)
c        print*,'sum_wab(i) ',sum_wab(i)
c        print*,'pVol(i) ',pVol(i)
c        if(i.lt.nbp1)then
c          print*,'xnb, znb    ',xnb(i),znb(i)
c          print*,'xtb, ztb    ',xtb(i),ztb(i)
c        endif
c        print*,'nb_local ',nb_local
c        print*,'np_local ',np_local
c        print*,'nbf_image_local       = ',nbf_image_local
c        print*,'nbfm_image_local = ',nbfm_image_local
c        print*,'nbfreem_image_local     = ',nbfreem_image_local
c        print*,'nb_image_local ',nb_image_local
c        print*,'np_image_local ',np_image_local
c        if(1.eq.2)then
c          !-- Image Print-out --
c          do ipcount=1,np_local
c            if(ipcount.le.nbf_local)then
c              i = ip_index_BP_local_1(ipcount)
c            else if(ipcount.le.nbfm_local)then
c              i = ip_index_BP_local_2(ipcount-nbf_local)
c            else if(ipcount.le.nbfreem_local)then
c              i = ip_index_BP_local_3(ipcount-nbfm_local)
c            else
c              i = ip_index_FP_local(ipcount-nb_local)
c            end if
c            print*,'ipcount -> i, Type, xp, zp',
c     &                  ipcount,i,i_particleType(i),xp(i),zp(i)
c          end do
c        endif
c       print*
c      end if

c     --- Primary Divide for Resident Fluid Particles only ---
       i_resident_0 = 1  !Resident Moving Particles
       call check_limits_2D      
       call ini_divide(2)
       call divide(nbp1_local,np_local,2,i_resident_0)   !3rd argument = kind_p)  

c      --- Primary Divide for Resident Moving Boundary Particles only ---
       !call recover_list   !Why here nad not below? (see step_predictor_corrector_MPI_2D.f)
       if(iopt_movingObject.eq.1)then
         call recover_list
         call divide(nbfp1_local,nb_local,1,i_resident_0)   !3rd argument = kind_p)
       endif

      call divide_ParallelPrep(1)   !argument 1 selects densityFilter option

      
      call ac          


c
c  ... compute corrections to:
c         (a) rate of change of velocity due to
c             (i)  body forces, and
c             (ii) boundary forces
c
c         (b) rate of change of position (free-surface, XSPH)
c
C
      call correct


c	To calculate time step 
      if (ivar_dt.eq.1) then
	    call variable_time_step  
	    ddt_p=dt_new
      endif
	

c
c  ... predictor
c

      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
         xo(i)   = xp(i)
c        yo(i)   =         
         zo(i)   = zp(i)
         uo(i)   = up(i)
c        vo(i)   =          
         wo(i)   = wp(i)
         po(i)   = p(i)
         rhoo(i) = rhop(i)
         TEo(i)  = TEp(i)

         Volo(i)       = pVol(i)
         Volrhoo(i)    = Volrho(i)
         Volrhouo(i)   = Volrhou(i)
c        Volrhovo(i)   = 
         Volrhowo(i)   = Volrhow(i)
         VolTotalEo(i)  = VolTotalE(i)
         
      enddo

      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
c  ...   Compute volume and density
         pVol(i) = Volo(i) + dt2*aVol(i)
         Volrho(i) = Volrhoo(i) + dt2*ar(i)
         rhop(i) = Volrho(i) / pVol(i)
      end do
      
      do ipcount=nb_local+1,np_local
         i = ip_index_FP_local(ipcount-nb_local)
         
         xp(i) = xo(i) + dt2*xdot(i)
c        yp(i) =
         zp(i) = zo(i) + dt2*zdot(i)

         Volrhou(i) = Volrhouo(i) + dt2*ax(i)
c        Volrhov(i) =         
         Volrhow(i) = Volrhowo(i) + dt2*az(i)
         
         up(i) = Volrhou(i) /Volrho(i)
c        vp(i) =
         wp(i) = Volrhow(i) /Volrho(i)
      
      enddo
      
      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
c  ...   Compute Total and thermal energy
         VolTotalE(i) = VolTotalEo(i) + dt2*aTE(i)
         TotalE(i) = VolTotalE(i) / pVol(i)
         TEp(i) = (TotalE(i)/rhop(i))
     &           - 0.5*(up(i)*up(i) + wp(i)*wp(i))
c  ...   compute pressure using the equation of state
         call equation_of_state(rhop(i),TEp(i),p(i),cs(i))
      enddo
            
c      !-- Debug Code------
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'AFTER CORRECT, PREDICTOR, before movingOBJECTS'
c        print*,'rank_id, itime ',rank_id, itime
c        i = i_ParticleCheck  
c        print*,'i = ',i
c        print*,'xp, zp  ',xp(i),zp(i)
c        print*,'up, wp  ',up(i),wp(i)
c        print*,'rhop, p,  pVol, pm ',rhop(i),p(i) ,pVol(i),pm(i)
c        print*,'rhoo, po, Volo, pm ',rhoo(i),po(i),Volo(i),pm(i)
c        print*,'VolTotalE   , VolTotalEo ',
c     &          VolTotalE(i), VolTotalEo(i) 
c        print*,'aTE(i) ',aTE(i)
c        print*,'TEp, TotalE ',TEp(i),TotalE(i)
c        print*,'ar(i), aVol(i) ',ar(i), aVol(i)
c        print*,'Volrhoo(i), Volrho(i) ',Volrhoo(i), Volrho(i) 
c        print*,'ax(i), Volrhou(i) ',ax(i), Volrhou(i)
c
c        print*,'az(i), Volrhow(i) ',az(i), Volrhow(i)
c        print*,'pVol(i), Volo(i) ',pVol(i), Volo(i)
c        print*,'dudx_CSPH(i)',dudx_CSPH(i)
c        print*,'dudz_CSPH(i)',dudz_CSPH(i)
c        print*,'dwdx_CSPH(i)',dwdx_CSPH(i)
c        print*,'dwdz_CSPH(i)',dwdz_CSPH(i)
c        print*,'drhodx_CSPH(i)',drhodx_CSPH(i)
c
c        print*,'drhodz_CSPH(i)',drhodz_CSPH(i)
c        print*,'sum_wab(i) ',sum_wab(i)
c        write(*,859)'ax(i) ',ax(i)
c        write(*,859)'az(i) ',az(i)
c
c       if(1.eq.2)then 
c        print*
c        i = i_ParticleCheck  
c        print*,'i = ',i
c        print*,'xp, zp  ',xp(i),zp(i)
c        print*,'up, wp  ',up(i),wp(i)
c        print*,'rhop, p,  pVol, pm ',rhop(i),p(i) ,pVol(i),pm(i)
c        print*,'rhoo, po, Volo, pm ',rhoo(i),po(i),Volo(i),pm(i)
c        print*,'VolTotalE   , VolTotalEo ',
c     &          VolTotalE(i), VolTotalEo(i) 
c        print*,'aTE(i) ',aTE(i)
c        print*,'TEp, TotalE ',TEp(i),TotalE(i)
c        print*,'ar(i), aVol(i) ',ar(i), aVol(i)
c        print*,'Volrhoo(i), Volrho(i) ',Volrhoo(i), Volrho(i) 
c        print*,'ax(i), Volrhou(i) ',ax(i), Volrhou(i)
c
c        print*,'az(i), Volrhow(i) ',az(i), Volrhow(i)
c        print*,'pVol(i), Volo(i) ',pVol(i), Volo(i)
c       end if
c       !stop
c      end if
c      !---------------------

      call movingObjects(0)
      if(iBC.eq.1.and.iopt_FloatingBodies.eq.1)then   !Repulsive Force BC
        i_particleTypeChoice = 3   !Free-moving BPs
        call updateNormals_2D(nbfreem_local+1,nb_local,
     &                          i_particleTypeChoice)   
      end if

c
c  ...  corrector
c
       nb_image_local      = nb_local
       np_image_local      = np_local
       nbf_image_local     = nbf_local
       nbfm_image_local    = nbfm_local
       nbfreem_image_local = nbfreem_local
       nbp1_local  = nb_local + 1
       nfbp1_local = nbf_local + 1     

c     --- Primary Divide for Resident Fluid Particles only ---
       i_resident_0 = 1  !Resident Moving Particles
       call check_limits_2D      
       call ini_divide(2)
       call divide(nbp1_local,np_local,2,i_resident_0)   !3rd argument = kind_p)  

c      --- Primary Divide for Resident Moving Boundary Particles only ---
       !call recover_list   !Why here nad not below? (see step_predictor_corrector_MPI_2D.f)
       if(iopt_movingObject.eq.1)then
         call recover_list
         call divide(nbfp1_local,nb_local,1,i_resident_0)   !3rd argument = kind_p)
       endif

      call divide_ParallelPrep(0)   !argument 1 selects densityFilter option
      
      
      call ac
      call correct


      if (ivar_dt.eq.1) then
          call variable_time_step  
          ddt_c=dt_new
          dt=min(ddt_p,ddt_c)  !Time step to be used in next loop
      endif

      
      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
         pVol(i) = Volo(i) + dt2*aVol(i)
         Volrho(i) = Volrhoo(i) + dt2*ar(i)
c  ...   Compute Total energy
         VolTotalE(i) = VolTotalEo(i) + dt2*aTE(i)
      end do

      
      
      do ipcount=nb_local+1,np_local
         i = ip_index_FP_local(ipcount-nb_local)

         xp(i) = xo(i) + dt2*xdot(i)
c        yp(i) =
         zp(i) = zo(i) + dt2*zdot(i)

         Volrhou(i) = Volrhouo(i) + dt2*ax(i)
c        Volrhov(i) =
         Volrhow(i) = Volrhowo(i) + dt2*az(i)
      enddo

      
      !-- Perform final integration correction --
      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
         pVol(i)   = 2.*pVol(i) - Volo(i)
         Volrho(i) = 2.*Volrho(i) - Volrhoo(i)
         rhop(i) = Volrho(i) / pVol(i)
      end do
      
      do ipcount=nb_local+1,np_local
         i = ip_index_FP_local(ipcount-nb_local)
         
         xp(i)   = 2.*xp(i) - xo(i)
c        yp(i) =
         zp(i)   = 2.*zp(i) - zo(i)
         
         Volrhou(i)   = 2.*Volrhou(i) - Volrhouo(i)
c        Volrhov(i) =
         Volrhow(i)   = 2.*Volrhow(i) - Volrhowo(i)
         
         up(i) = Volrhou(i) / Volrho(i)
c        vp(i) = 
         wp(i) = Volrhow(i) / Volrho(i)
      
      enddo

      do ipcount=nstart,np_local
         call ipcount_to_i(ipcount,i)
c  ...   Compute thermal energy
         VolTotalE(i)=2.*VolTotalE(i)-VolTotalEo(i)
         TotalE(i) = VolTotalE(i) / pVol(i)
         TEp(i) = TotalE(i)/rhop(i)
     &          - 0.5*(up(i)*up(i) + wp(i)*wp(i))
c  ...   compute pressure using the equation of state
         call equation_of_state(rhop(i),TEp(i),p(i),cs(i))
      enddo

       
      call movingObjects(1)
      
c      !- Diagnostic Info -
c      if(itime.ge.itime_check.and.rank_id.eq.0)then
c        print*
c        print*,'In STEP_PREDICTOR_CORRECTOR.f before INDEX_CHECK.f'
c        i =  i_ParticleCheck
c        print*,'i, xp, zp ',i,xp(i),zp(i)
c        print*,'i, up, wp ',i,up(i),wp(i)
c        print*,'rhop(i) ',rhop(i)
c        print*,'p(i)    ',p(i)
c        print*,'sum_wab(i) ',sum_wab(i)
c        print*
c      end if
c      !stop
      
      
      call index_check_EastWest
      if(i_call_UpDown    .eq.1) call index_check_UpDown
c     if(i_call_NorthSouth.eq.1) call index_check_NorthSouth  !3-D only


c     Normals are calculated in subroutine updateNormals
      if(iBC.eq.1)then   !Repulsive Force BC
       i_particleTypeChoice = 0
       if(i_MPI_Adapt.eq.1)then
         nbf_uN_start = 1             !All BPs
       else
         nbf_uN_start = nbf_local+1   !All moving BPs
       endif
       call updateNormals_2D(nbf_uN_start,nb_local,i_particleTypeChoice)   
      end if      
      
      call MPI_BARRIER(MPI_COMM_WORLD, ierr)   !Synchronise all processes to this point
      
c      if(itime.ge.itime_check.and.rank_id.eq.i_rank_check)then
c        print*
c        print*,'Accn Check, AFTER CORRECT, CORRECTOR'
c        print*,'rank_id, itime ',rank_id, itime
c        i = i_ParticleCheck  
c        print*,'i ',i
c        print*,'xp, zp  ',xp(i),zp(i)
c        print*,'up, wp  ',up(i),wp(i)
c        print*,'rhop, p,  pVol, pm ',rhop(i),p(i) ,pVol(i),pm(i)
c        print*,'rhoo, po, Volo, pm ',rhoo(i),po(i),Volo(i),pm(i)
c        print*,'VolTotalE   , VolTotalEo ',
c     &          VolTotalE(i), VolTotalEo(i) 
c        print*,'aTE(i) ',aTE(i)
c        print*,'TEp, TotalE ',TEp(i),TotalE(i)
c        print*,'ar(i), aVol(i) ',ar(i), aVol(i)
c        print*,'Volrhoo(i), Volrho(i) ',Volrhoo(i), Volrho(i) 
c        print*,'ax(i), Volrhou(i) ',ax(i), Volrhou(i)
c
c        print*,'az(i), Volrhow(i) ',az(i), Volrhow(i)
c        print*,'pVol(i), Volo(i) ',pVol(i), Volo(i)
c        print*,'dudx_CSPH(i)',dudx_CSPH(i)
c        print*,'dudz_CSPH(i)',dudz_CSPH(i)
c        print*,'dwdx_CSPH(i)',dwdx_CSPH(i)
c        print*,'dwdz_CSPH(i)',dwdz_CSPH(i)
c        print*,'drhodx_CSPH(i)',drhodx_CSPH(i)
c
c        print*,'drhodz_CSPH(i)',drhodz_CSPH(i)
c        print*,'sum_wab(i) ',sum_wab(i)
c        write(*,859)'ax(i) ',ax(i)
c
c        write(*,859)'az(i) ',az(i)
859     format(a11,' ',e24.16)        
c        if(1.eq.2)then
c         i = 8822   !469
c         print*
c         print*,'i = ',i
c         print*,'xp, zp  ',xp(i),zp(i)
c         print*,'up, wp  ',up(i),wp(i)
c         print*,'rhop, p,  pVol, pm ',rhop(i),p(i) ,pVol(i),pm(i)
c         print*,'rhoo, po, Volo, pm ',rhoo(i),po(i),Volo(i),pm(i)
c         print*,'VolTotalE   , VolTotalEo ',
c     &           VolTotalE(i), VolTotalEo(i) 
c         print*,'aTE(i) ',aTE(i)
c         print*,'TEp, TotalE ',TEp(i),TotalE(i)
c         print*,'ar(i), aVol(i) ',ar(i), aVol(i)
c         print*,'Volrhoo(i), Volrho(i) ',Volrhoo(i), Volrho(i) 
c         print*,'ax(i), Volrhou(i) ',ax(i), Volrhou(i)
c         print*,'pVol(i), Volo(i) ',pVol(i), Volo(i)
c        endif
c        if(itime.eq.1)stop
c        !stop
c      end if
      
      
      end
c