c    "Copyright 2010 Prof. Robert Dalrymple, Prof. M. Gomez Gesteira, Dr Benedict Rogers, 
c     Dr Alejandro Crespo, Dr Muthukumar Narayanaswamy, Dr Shan Zou, Dr Andrea Panizzo "
c
c    This file is part of SPHYSICS.
c
c    SPHYSICS is free software; you can redistribute it and/or modify
c    it under the terms of the GNU General Public License as published by
c    the Free Software Foundation; either version 3 of the License, or
c    (at your option) any later version.
c
c    SPHYSICS is distributed in the hope that it will be useful,
c    but WITHOUT ANY WARRANTY; without even the implied warranty of
c    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c    GNU General Public License for more details.
c
c    You should have received a copy of the GNU General Public License
c    along with this program.  If not, see <http://www.gnu.org/licenses/>.

      subroutine updateNormals_2D(i_start,i_stop,i_particleTypeChoice)

      include 'common.2D'
      
      double precision dxpb1,dxpb2
      
c       Subroutine calculates tangents and normals for a surface given 
c       by three points:
c
c       Vectors t, s, n form a righ-handed (orthogonal) system
c
c
c       (xi     , yi    , zi )    = point of interest 
c       (xt_next,yt_next,zt_next) = next point in traverse of plane
c       (xs_next,ys_next,zs_next) = corresponding point in orthogonal traverse

!               ^ (drs)
!              /
!            
!       ^(n) o(3)
!       |   /
!       |  /
!       | /
!       |/ )
!    (i)o---------o(2)  -->(drt)
c    
c      i_start = 1         !BedPs only
c      i_stop  = nbf       !Bed particles only
c
c      i_start = nbf + 1   !Moving Solid Object particles only
c      i_stop  = nb        !Moving Solid Object particles only
c
c       Specifically written for use in Monaghan's BC

    
      !-- MPI Version --
      do ipcount=i_start,i_stop  
       call ipcount_to_i(ipcount,i)
       
       if(i.lt.1)then
         print*
         print*,'ERROR in updateNormals_2D'
         print*,'rank_id, itime  ',rank_id, itime
         print*,'i_start, i_stop ',i_start,i_stop
         print*,'nbf_local       ' ,nbf_local 
         print*,'nbfm_local      ' ,nbfm_local 
         print*,'nbfreem_local   ' ,nbfreem_local 
         print*
         stop
       endif
         
        xp_i = xp(i)
c       yp_i =
        zp_i = zp(i)
        
        i_absolute = iabsolute_BP_Pointer_Index(i)
        iminus1_neighbour = iBP_Pointer_Info(i_absolute,3)
        iplus1_neighbour  = iBP_Pointer_Info(i_absolute,4)
c       jminus1_neighbour = 
c       jplus1_neighbour  = 
           
        xp_BP_iminus1 = BP_xyz_Data(iminus1_neighbour,1)
        zp_BP_iminus1 = BP_xyz_Data(iminus1_neighbour,2)
        xp_BP_iplus1  = BP_xyz_Data(iplus1_neighbour ,1)
        zp_BP_iplus1  = BP_xyz_Data(iplus1_neighbour ,2)

        
        if(i_absolute.eq.iminus1_neighbour)then
          xp_previous = xp_i - (xp_BP_iplus1-xp_i)
          zp_previous = zp_i - (zp_BP_iplus1-zp_i)
          xp_next = xp_BP_iplus1
          zp_next = zp_BP_iplus1
        else if(i_absolute.eq.iplus1_neighbour)then
          xp_previous = xp_BP_iminus1
          zp_previous = zp_BP_iminus1
          xp_next = xp_i + (xp_i-xp_BP_iminus1)
          zp_next = zp_i + (zp_i-zp_BP_iminus1)
        else
          xp_previous = xp_BP_iminus1
          zp_previous = zp_BP_iminus1
          xp_next = xp_BP_iplus1
          zp_next = zp_BP_iplus1
        end if
        
        dxn_forward=(xp_next-xp_i)
        dzn_forward=(zp_next-zp_i)
        dxn_back=(xp_i-xp_previous)
        dzn_back=(zp_i-zp_previous)
        
        if(i_periodicOBs(1).eq.1)then
          if(dxn_forward.gt.six_h)then
            dxpb1 = xmax_container_double-dble(xp_next)
            dxpb2 = dble(xp_i)-xmin_container_double
            dxn_forward = -real(dxpb1+dxpb2)
          else if(dxn_forward.lt.-six_h)then
            dxpb1 = xmax_container_double-dble(xp_i)
            dxpb2 = dble(xp_next)-xmin_container_double
            dxn_forward = real(dxpb1+dxpb2)
          end if
          if(dxn_back.gt.six_h)then
            dxpb1 = xmax_container_double-dble(xp_i)
            dxpb2 = dble(xp_previous)-xmin_container_double
            dxn_back = -real(dxpb1+dxpb2)
          else if(dxn_back.lt.-six_h)then
            dxpb1 = xmax_container_double-dble(xp_previous)
            dxpb2 = dble(xp_i)-xmin_container_double
            dxn_back = real(dxpb1+dxpb2)
          end if
        end if
        
        rbn_back = sqrt(dxn_back*dxn_back + dzn_back*dzn_back)
        rbn_forward = sqrt(dxn_forward*dxn_forward +
     &                     dzn_forward*dzn_forward)
        dzn = dzn_forward + dzn_back
        dxn = dxn_forward + dxn_back
        rb=dxn*dxn+dzn*dzn
        xnb(i)=-dzn/sqrt(rb)
        znb(i)=dxn/sqrt(rb)
        xtb(i) =  znb(i)
        ztb(i) = -xnb(i)
        if(abs(rbn_back).gt.0.0)then
          deltaptb(i,1) = rbn_back
        else
          deltaptb(i,1) = h*1.0e-3
        end if          
        if(abs(rbn_forward).gt.0.0)then
          deltaptb(i,2) = rbn_forward
        else
          deltaptb(i,2) = h*1.0e-3
        end if 
        
c        if(i.eq.103.and.rank_id.eq.3)then
c          print*
c          print*,'i, ipcount ',i,ipcount
c          print*,'xp, zp ',xp(i),zp(i)
c          print*,'i_absolute ',i_absolute
c          print*,'iminus1_neighbour ',iminus1_neighbour
c          print*,'iplus1_neighbour ',iplus1_neighbour
c          print*,'i_start,i_stop ',i_start,i_stop
c          print*,'xp_BP_iminus1, zp_BP_iminus1 ',
c     &            xp_BP_iminus1, zp_BP_iminus1
c          print*,'xp_BP_iplus1 , zp_BP_iplus1 ',
c     &            xp_BP_iplus1 , zp_BP_iplus1
c          print*,'xnb, znb ',xnb(i),znb(i)
c          print*,'xtb, ztb ',xtb(i),ztb(i)
c          print*,'deltaptb(i,1&2)',deltaptb(i,1),deltaptb(i,2)
c          print*
c        endif

       !endif   !End of:   if(i_particleType(i).eq.i_particleTypeChoice)then
      
      enddo


      end subroutine
